import wx


def throw_error(message, title='Error', parent=None):
    """
    Generic way to throw an error and display the appropriate dialog
    """
    dlg = wx.MessageDialog(parent, message, title, wx.OK | wx.ICON_ERROR)
    dlg.CenterOnParent()
    dlg.ShowModal()
    dlg.Destroy()



import os
import pydicom
import pytest
import time

from queue import Queue

from dicomsort.dicomsorter import Dicom, DicomSorter, Sorter
from dicomsort.errors import DicomFolderError


def default_sorter():
    return Sorter(Queue(), '', [], '')


class TestDicom:
    def test_constructor_without_dicom(self, dicom_generator, mocker):
        func = mocker.patch.object(pydicom, 'read_file', return_value='dicom')

        filename, _ = dicom_generator()
        dcm = Dicom(filename)

        func.assert_called_once_with(filename)
        assert dcm.dicom == 'dicom'

    def test_constructor_with_dicom(self, dicom_generator, mocker):
        func = mocker.patch.object(pydicom, 'read_file', return_value='dicom')

        filename, dataset = dicom_generator()

        dcm = Dicom(filename, dcm=dataset)

        func.assert_not_called()
        assert dcm.dicom == dataset

    def test_get_item_override_function(self, dicom_generator):
        filename, dicom = dicom_generator()
        dcm = Dicom(filename, dcm=dicom)

        dcm.set_anonymization_rules({'Field': lambda: 'Value'})

        assert dcm['Field'] == 'Value'

    def test_get_item_override_value(self, dicom_generator):
        filename, dicom = dicom_generator()
        dcm = Dicom(filename, dcm=dicom)

        dcm.set_anonymization_rules({'Field': 'Value'})

        assert dcm['Field'] == 'Value'

    def test_get_item_dicom_attribute(self, dicom_generator):
        filename, dicom = dicom_generator(PatientName='Suever')
        dcm = Dicom(filename, dcm=dicom)

        assert dcm['PatientName'] == 'Suever'

    def test_get_file_extension(self, dicom_generator):
        filename, _ = dicom_generator('image.ext')

        dcm = Dicom(filename)
        assert dcm._file_extension() == '.ext'

    def test_get_file_extension_empty(self, dicom_generator):
        filename, _ = dicom_generator('image')

        dcm = Dicom(filename)
        assert dcm._file_extension() == ''

    def test_get_series_description(self, dicom_generator):
        filename, dataset = dicom_generator(
            SeriesDescription=' My Series',
            SeriesNumber=1
        )

        dcm = Dicom(filename, dcm=dataset)

        assert dcm._series_description() == 'My Series_Series0001'

    def test_get_series_description_no_description(self, dicom_generator):
        filename, dicom = dicom_generator(SeriesNumber=1)
        del dicom.SeriesDescription

        dcm = Dicom(filename, dcm=dicom)

        assert dcm._series_description() == 'Series0001'

    def test_get_series_description_series_first(self, dicom_generator):
        filename, dataset = dicom_generator(
            SeriesDescription='My Series',
            SeriesNumber=1
        )

        dcm = Dicom(filename, dcm=dataset)
        dcm.series_first = True

        assert dcm._series_description() == 'Series0001_My Series'

    def test_get_patient_age_with_field(self, dicom_generator):
        age = '10Y'
        filename, dataset = dicom_generator(PatientAge=age)

        dcm = Dicom(filename, dcm=dataset)

        assert dcm._patient_age() == age

    def test_get_patient_age_without_birth_date(self, dicom_generator):
        filename, dataset = dicom_generator()

        dcm = Dicom(filename, dcm=dataset)

        assert dcm._patient_age() == ''

    def test_get_patient_age_with_empty_birth_date(self, dicom_generator):
        filename, dataset = dicom_generator(PatientBirthDate='')

        dcm = Dicom(filename, dcm=dataset)

        assert dcm._patient_age() == ''

    def test_get_patient_age_with_birth_date(self, dicom_generator):
        filename, dataset = dicom_generator(
            PatientBirthDate='20180101',
            StudyDate='20190701'
        )

        dcm = Dicom(filename, dcm=dataset)

        assert dcm._patient_age() == '001Y'

    def test_get_image_type_no_type(self, dicom_generator):
        filename, dataset = dicom_generator()

        dcm = Dicom(filename, dcm=dataset)

        assert dcm._image_type() == 'Unknown'

    def test_get_image_type_magnitude(self, dicom_generator):
        filename, dataset = dicom_generator(ImageType=['FFE', 'M', 'JUNK'])

        dcm = Dicom(filename, dcm=dataset)

        assert dcm._image_type() == 'Mag'

    def test_get_image_type_generic(self, dicom_generator):
        filename, dataset = dicom_generator(ImageType=['A', 'B'])

        dcm = Dicom(filename, dcm=dataset)

        assert dcm._image_type() == 'Image'

    def test_get_image_type_phase(self, dicom_generator):
        filename, dataset = dicom_generator(ImageType=['P', 'BLAH'])

        dcm = Dicom(filename, dcm=dataset)

        assert dcm._image_type() == 'Phase'

    def test_get_image_type_phoenix(self, dicom_generator):
        filename, dataset = dicom_generator(ImageType=['CSA REPORT', ''])

        dcm = Dicom(filename, dcm=dataset)

        assert dcm._image_type() == 'Phoenix'

    def test_get_image_type_3drecon(self, dicom_generator):
        filename, dataset = dicom_generator(ImageType=['CSA 3D EDITOR', ''])

        dcm = Dicom(filename, dcm=dataset)

        assert dcm._image_type() == '3DRecon'
        assert dataset.InstanceNumber == dataset.SeriesNumber

    def test_create_default_overrides(self, dicom_generator):
        filename, dataset = dicom_generator()
        dcm = Dicom(filename, dcm=dataset)

        assert 'ImageType' in dcm.default_overrides
        assert 'FileExtension' in dcm.default_overrides
        assert 'SeriesDescription' in dcm.default_overrides

        assert dcm.overrides == dcm.default_overrides

    def test_is_anonymous(self, dicom_generator):
        filename, dataset = dicom_generator()
        dcm = Dicom(filename, dcm=dataset)

        assert dcm.is_anonymous() is False

        # Set anonimization rules
        dcm.set_anonymization_rules({'PatientName': 'ANON'})

        assert dcm.is_anonymous() is True

    def test_anonymization_invalid_input(self, dicom_generator):
        filename, dataset = dicom_generator()
        dcm = Dicom(filename, dcm=dataset)

        with pytest.raises(Exception) as excinfo:
            dcm.set_anonymization_rules('')

        assert excinfo.value.args[0] == 'Anon rules must be a dictionary'

    def test_anonymization(self, dicom_generator):
        filename, dataset = dicom_generator()
        dcm = Dicom(filename, dcm=dataset)

        adict = {'Key': 'Value'}

        dcm.set_anonymization_rules(adict)

        assert dcm.anonymization_lookup == adict
        assert dcm.overrides['Key'] == 'Value'
        assert dcm['Key'] == 'Value'

    def test_anonymize_age(self, dicom_generator):
        filename, dataset = dicom_generator(
            PatientAge='030Y',
            PatientBirthDate='19890201',
            StudyDate='20190301'
        )

        dcm = Dicom(filename, dcm=dataset)

        dcm.set_anonymization_rules({'PatientBirthDate': ''})

        assert dcm['PatientAge'] == '030Y'
        assert dcm['PatientBirthDate'] == '19890101'

    def test_anonymize_birthdate(self, dicom_generator):
        date = '20191101'
        filename, dataset = dicom_generator(PatientBirthDate='20200101')
        dcm = Dicom(filename, dcm=dataset)

        dcm.set_anonymization_rules({'PatientBirthDate': date})

        assert dcm.overrides['PatientBirthDate'] == date
        assert dcm['PatientBirthDate'] == date

    def test_anonymize_empty_birthdate(self, dicom_generator):
        filename, dataset = dicom_generator(
            PatientBirthDate='20170601',
            StudyDate='20180201',
        )
        dcm = Dicom(filename, dcm=dataset)

        dcm.set_anonymization_rules({'PatientBirthDate': ''})

        # Maintains the patient age but uses January 01
        assert dcm.overrides['PatientBirthDate'] == '20180101'
        assert dcm['PatientBirthDate'] == '20180101'

        # 1-day Old
        dataset.PatientBirthDate = '20180131'

        dcm.set_anonymization_rules({'PatientBirthDate': ''})
        assert dcm.overrides['PatientBirthDate'] == '20180101'
        assert dcm['PatientBirthDate'] == '20180101'

    def test_get_destination(self, dicom_generator):
        filename, dicom = dicom_generator(
            PatientName='name',
            SeriesDescription='desc',
            SeriesNumber=1,
        )

        dcm = Dicom(filename, dcm=dicom)

        # Base Directory
        root = '/my/base/directory'

        # Sub-directories
        directory = [
            '%(ImageType)s',
            '%(SeriesDescription)s %(PatientName)s'
        ]

        # Filename format
        filename = '%(SeriesNumber)d file'

        dest = dcm.get_destination(root, directory, filename)

        expected = '/my/base/directory/Unknown/desc_Series0001 name/1 file'
        assert dest == expected

    def test_get_destination_bad_filename_attr(self, dicom_generator):
        filename, dicom = dicom_generator(
            'image.dcm',
            PatientName='name',
            SeriesDescription='desc',
            SeriesNumber=1,
        )

        dcm = Dicom(filename, dcm=dicom)

        # Base Directory
        root = '/my/base/directory'

        # Sub-directories
        directory = [
            '%(ImageType)s',
            '%(SeriesDescription)s %(PatientName)s'
        ]

        # Filename format which uses attributes that are not defined
        filename = '%(Invalid)d file'

        dest = dcm.get_destination(root, directory, filename)
        expected = '/my/base/directory/Unknown/desc_Series0001 name/image.dcm'

        assert dest == expected

    def test_get_destination_bad_directory_attr(self, dicom_generator):
        filename, dicom = dicom_generator(
            'image.dcm',
            PatientName='name',
            SeriesDescription='desc',
            SeriesNumber=1,
        )

        dcm = Dicom(filename, dcm=dicom)

        # Base Directory
        root = '/my/base/directory'

        # Sub-directories
        directory = [
            '%(Invalid)s',  # Invalid attribute
            '%(SeriesDescription)s %(PatientName)s'
        ]

        # Filename format
        filename = '%(SeriesNumber)d file'

        dest = dcm.get_destination(root, directory, filename)

        expected = '/my/base/directory/UNKNOWN/desc_Series0001 name/1 file'
        assert dest == expected

    def test_sort_inplace(self, dicom_generator, tmpdir):
        filename, dicom = dicom_generator(
            'image.dcm',
            SeriesDescription='desc',
            SeriesNumber=1,
        )
        dcm = Dicom(filename, dcm=dicom)

        # Base Directory
        root = str(tmpdir)

        # Subdirectories
        directory = None

        # Filename format
        file_format = '%(ImageType)s'

        dcm.sort(
            root, directory, file_format,
            rootdir=[os.path.dirname(filename)],
        )

        destination = str(tmpdir.join('image.dcm'))

        assert os.path.exists(destination)

    def test_sort_keep_original(self, dicom_generator, tmpdir):
        filename, dicom = dicom_generator(
            SeriesDescription='desc',
            SeriesNumber=1,
        )
        dcm = Dicom(filename, dcm=dicom)

        # Base Directory
        root = str(tmpdir)

        # Subdirectories
        directory = [
            '%(SeriesDescription)s',
            '%(SeriesDescription)s',
        ]

        # Filename format
        file_format = '%(ImageType)s'

        dcm.sort(root, directory, file_format, keep_original=True)

        destination = tmpdir.join('desc_Series0001').join('desc_Series0001')
        destination = destination.join('Unknown')

        assert os.path.exists(filename)
        assert os.path.exists(str(destination))

    def test_sort_discard_original(self, dicom_generator, tmpdir):
        filename, dicom = dicom_generator(
            SeriesDescription='desc',
            SeriesNumber=1,
        )
        dcm = Dicom(filename, dcm=dicom)

        # Base Directory
        root = str(tmpdir)

        # Subdirectories
        directory = [
            '%(SeriesDescription)s',
            '%(SeriesDescription)s',
        ]

        # Filename format
        file_format = '%(ImageType)s'

        dcm.sort(root, directory, file_format, keep_original=False)

        destination = tmpdir.join('desc_Series0001').join('desc_Series0001')
        destination = destination.join('Unknown')

        assert os.path.exists(filename) is False
        assert os.path.exists(str(destination))

    def test_sort_anonymize(self, dicom_generator, tmpdir):
        filename, dicom = dicom_generator(
            PatientName='TO^BE^REMOVED',
            SeriesDescription='desc',
            SeriesNumber=1,
        )
        dcm = Dicom(filename, dcm=dicom)

        # Overwrite the PatientName
        dcm.set_anonymization_rules({'PatientName': 'ANON'})

        # Base Directory
        root = str(tmpdir)

        # Subdirectories
        directory = [
            '%(PatientName)s',
            '%(SeriesDescription)s',
        ]

        # Filename format
        file_format = '%(ImageType)s'

        dcm.sort(root, directory, file_format, keep_original=False)

        destination = tmpdir.join('ANON').join('desc_Series0001')
        destination = str(destination.join('Unknown'))

        assert os.path.exists(destination)

        # Ensure the DICOM was modified
        newdcm = pydicom.read_file(destination)

        assert newdcm.PatientName == 'ANON'

    def test_sort_anonymize_invalid_field(self, dicom_generator, tmpdir):
        filename, dicom = dicom_generator(
            PatientName='TO^BE^REMOVED',
            SeriesDescription='desc',
            SeriesNumber=1,
        )
        dcm = Dicom(filename, dcm=dicom)

        # Overwrite the PatientID (this field does not exist)
        dcm.set_anonymization_rules({'PatientID': 'ANON'})

        # Base Directory
        root = str(tmpdir)

        # Subdirectories
        directory = [
            '%(PatientID)s',
            '%(SeriesDescription)s',
        ]

        # Filename format
        file_format = '%(ImageType)s'

        dcm.sort(root, directory, file_format, keep_original=False)

        destination = tmpdir.join('ANON').join('desc_Series0001')
        destination = str(destination.join('Unknown'))

        assert os.path.exists(destination)

        # Make sure the field was not added to the new DICOM
        newdcm = pydicom.read_file(destination)

        assert 'PatientID' not in newdcm

    def test_sort_test(self, dicom_generator, tmpdir, capsys):
        filename, dicom = dicom_generator(
            SeriesDescription='desc',
            SeriesNumber=1,
        )
        dcm = Dicom(filename, dcm=dicom)

        # Base Directory
        root = str(tmpdir)

        # Subdirectories
        directory = [
            '%(SeriesDescription)s',
            '%(SeriesDescription)s',
        ]

        # Filename format
        file_format = '%(ImageType)s'

        dcm.sort(root, directory, file_format, keep_original=False, test=True)

        destination = tmpdir.join('desc_Series0001').join('desc_Series0001')
        destination = destination.join('Unknown')

        # The original is unmodified
        assert os.path.exists(filename)

        # The file was not actually copied
        assert os.path.exists(str(destination)) is False

        captured = capsys.readouterr()
        assert captured.out == str(destination) + '\n'


class TestDicomSorter:
    def test_constructor_defaults(self):
        sorter = DicomSorter()

        assert sorter.pathname == [os.getcwd(), ]

        # Booleans
        assert sorter.keep_filename is False
        assert sorter.series_first is False
        assert sorter.keep_original is True

        assert sorter.sorters == []
        assert sorter.folders == []
        assert sorter.anonymization_lookup == dict()

    def test_constructor_single_path(self):
        path = '/path'
        sorter = DicomSorter(path)

        assert sorter.pathname == [path, ]

    def test_constructor_multiple_paths(self):
        paths = ['/path1', '/path2']
        sorter = DicomSorter(paths)

        assert sorter.pathname == paths

    def test_get_available_fields_no_dicoms(self, tmpdir):
        fobj = tmpdir.join('tmp')
        fobj.write('invalid')

        root = str(tmpdir)

        sorter = DicomSorter(root)

        with pytest.raises(DicomFolderError):
            sorter.available_fields()

    def test_get_available_fields(self, dicom_generator):
        filename, dicom = dicom_generator()
        expected = dicom.dir('')

        root = os.path.dirname(filename)

        sorter = DicomSorter(root)

        assert sorter.available_fields() == expected

    def test_is_sorting_no_sorters(self):
        sorter = DicomSorter()

        assert sorter.sorters == []
        assert sorter.is_sorting() is False

    def test_is_sorting_active_sorters(self, mocker):
        sorter = DicomSorter()

        patch = mocker.patch.object(Sorter, 'is_alive')
        patch.return_value = True

        # Generate some sorters
        sorter.sorters.append(default_sorter())
        sorter.sorters.append(default_sorter())

        assert sorter.is_sorting() is True

    def test_is_sorting_inactive_sorters(self, mocker):
        sorter = DicomSorter()

        patch = mocker.patch.object(Sorter, 'is_alive')
        patch.return_value = False

        # Generate some sorters
        sorter.sorters.append(default_sorter())
        sorter.sorters.append(default_sorter())

        assert sorter.is_sorting() is False

    def test_get_folder_format_no_folders(self):
        sorter = DicomSorter()
        sorter.folders = None

        assert sorter.folder_format() is None

    def test_get_folder_format(self):
        sorter = DicomSorter()
        original = ['1', '2', '3']
        sorter.folders = original

        copy = sorter.folder_format()

        assert copy == original

        # modify the original to ensure it is a different object
        original.append('4')

        assert copy != original

    def test_set_anon_rules(self):
        sorter = DicomSorter()

        replacements = {
            'Key1': 'Value1',
            'Key2': 'Value2',
        }

        sorter.set_anonymization_rules(replacements)
        assert sorter.anonymization_lookup == replacements

    def test_set_anon_rules_invalid(self):
        sorter = DicomSorter()

        with pytest.raises(Exception) as exc_info:
            sorter.set_anonymization_rules('')

        assert exc_info.value.args[0] == 'Anon rules must be a dictionary'

    def test_sort(self, dicom_generator, mocker, tmpdir):
        filename, _ = dicom_generator(
            SeriesDescription='desc',
            SeriesNumber=1,
            InstanceNumber=1
        )

        sorter = DicomSorter()
        sorter.pathname = [os.path.dirname(filename), ]
        sorter.folders = [
            '%(SeriesDescription)s'
        ]

        output = tmpdir.join('output')

        sorter.sort(str(output))

        # Wait for sorting to complete
        while sorter.is_sorting():
            time.sleep(0.1)

        assert len(sorter.sorters) == 1

        series_folders = output.listdir()
        assert len(series_folders) == 1

        series_folder = series_folders[0]
        assert os.path.basename(str(series_folder)) == 'desc_Series0001'

        images = series_folder.listdir()
        assert len(images) == 1
        assert os.path.basename(str(images[0])) == 'Unknown (0001).dcm'



# ----------------------------------------------------------------------
# This file was generated by img2py.py
#
from wx.lib.embeddedimage import PyEmbeddedImage

main = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAABHNCSVQICAgIfAhkiAAACMdJ"
    "REFUWIW9l2tsHFcVx3/3zsyu17uxN7sbx5s0tus0ruskdR8CUqAVKSUJVH3xUCmlLSAFFYEQ"
    "FahIiEo8BBIV4kNB4lEBUikIBBKvQgt98khaQkPtxE5LUztmHdsbx16vvY/Zedx7+GASpTRV"
    "gxD8P525c3XO78w9c+aMwzmor6/vsXw+f+OmTZsu3r179/rp6enI9/3FM/d0d3dTr9fPxd3L"
    "pM5l08aNG18yxvQDSimFiJDJZABOtLW1jRYKhb9lMpmRW2+99dAtt9zy/P8EwFq7GUBEUEqR"
    "TqdptVoAkslk1MrKCp7nYa2lt7e3PDAwMLpv377RTZs2jd58882/3rt3b+1svt3/hNZai9b6"
    "Zba1ViUSCVzXRURwHIdSqdTd19e3PgzD3WNjYxSLxbcBj53Npz6XwCLKnspcRDDG0NnZied5"
    "ADiOQyaTwXVdrLU4jsPY2JgaHh5GREgmk/bVfL8qgD/3DHN9HRd1mZG5rkJn75n3isUiS0tL"
    "NJtNhoaGSKVShGGIiJDL5dBaE8cxMzMzZLNZ4jh+1eReFSBV3MHrH9779hMzc92OJnFqPZlM"
    "ks1mabVaeJ7H/Pw8tVqNbdu24fs+WmuMMYgIQRBw++138Jk7Nrzjhh/3X5qsjqcARp/86msD"
    "AHz/hw9vwBoAlFJ4nkd3dzfHjh0jn8/T1dVFPp+nv7+f4eFhcrkcy8vLpNNp4jims7OTF4++"
    "KPv/nvrkNy76wt/mnv5e88bLJn8+vPNT5wbw+J/+WrQmQLFadHfeeSdRFOF5Hlu2bGF4eJjB"
    "wUF6e3sZGRnh2muvRWtNsVjE8zyWlpbYtWu3qldnObliMPdP8JMf/ext53QEAI8+PbqZOMSK"
    "pVAo0NPTw/LyMlu3bsXzPEqlEocPH6ZUKjE+Pk4ulyOfzxOGIbVajVqtRro9TX3+eeqTj1D/"
    "eBetlZm0mfx65Rt3HXnTywCqEw+9AmCqNLNe4tUnsGfPHsbGxshmswwODlIul08XXHd3Nzt2"
    "7GBmZoZMJoMxhmKxiFKKRDJJEEYEYYQTGrARtflDa6efT50PoHcMR8989hNXPPDBWwt7r/zt"
    "L/vPBCiVyhtt1MLGMdu3b+fAgQPk83m01uRyOZrNJj09PYgIAwMDBEFwulbq9TpKKRxHE4Ux"
    "cWxQjmDFAAkWKt8pAei/PP3IJZ/7wj3v/+mj/ncef/PaiVr5IXnDYHnfxIHbPluvL3gS1Gn6"
    "PslkkqmpKdLpNMYYkskkk5OTpxvQ/Pw8mUyGcrnMyZMnKRQKaK0Jw5AwMpg4whEBYwGPe7/8"
    "8TkAff99H3kAYxVxHW2rJNsC/rzviSt6Nu/64poLcxIVFB/50E7aEz4iQqVS4ejRowA0m032"
    "79/P7Ows4+PjKKWoVCpks1lWVlZQSmGtJQwjxFq0ErCWyEL+muJLAPpz8/d93tLCEGIlQJkQ"
    "EGUiHxU2lOc1uXh7li0bjvHh269krnyCiYkJRkdH6ejoIJlMUi6X6ejoYHJy8nTWbW1tAMRR"
    "TGwsIDgaEENsFIAA6ONfe+fMe24cfkJJIMpGWBthpAUSIBKBjZGwwuJihUuG1vCB916B62oK"
    "69aTTmdoNBosLCwQxzGHDh063aZrtRqu67K4uEgU2dXgCMbGrFmTqZ6qM/fk7D1cv+GRe7Cd"
    "+6xYNDFiLDYOEWuQaJEoDGhFMSiHy4ZSvO7iqylNL3KsVGfsRcXU9DKl6RmCcLVHpNNpfN/H"
    "dV3iOCZo1XGiKrap0FrTu7GjdBpg3YYvAuzfNvi7F8aOTF8oYpQVg9gATIg1NYx4dGTXsrKy"
    "hLGWqF6jkDUU1ni8YetGvMRmGr6iUk/w3PgJjJNnZCTCGAMKwiAiDAKWq3W0A32F1PSpacYF"
    "mGz+gF3m0k9LY/KXiEGsRUyExFVQLogmmXBRYjGxj9YKYoV2PZRo/HoTJUIx47FlZ4FWENCR"
    "uorfPHoQ3/dRosFCw4/RGBk4f/3kwTM7YX/7bby0putX2c5ERSQWrQSxEcY0UNIGYkh4DlrF"
    "OFpjY4sYgyMOLd8HpXCUg1ihWlkmaLS4ariDgQsvZHZmFgBPK0xsieNY7dm1+/hZW/F7rnz8"
    "MyKREhNiwxpYhVhBxNKoVwn8FqlUO67jIAJhGOJpF09pMEIcRHiOi8QWamWuefN2FhYXVgMp"
    "MLEhjuC2933ghETySoAHnv3Rt11phsbUhKAC4mBtTL5DaNSriCiWl6qIEbIdnSS9BAnPw9UO"
    "WkBZIfBbiLXY2HJ+vsbsbBn9r7cujCGWBPd/tHhIeeqVAAD/ePYT98Ytq0wco7DYqMFQMeLy"
    "3gaX90X0r1dk2ixtySSudkklUyQTydW6EUAERBBj6Uq0SOgWrqNQQBwLRlze+t0fls96BBK8"
    "wOYbOu6Zm52d0gLaOijSYnFwccgkIjZ2NNja1aCvbYr+tcuscaqEzWUEjUaBtWAsrnaIWxE7"
    "BttRVsAKxliCWHPBXW9Z/SDwb0OpSg6eMs+Hru3tM9e/++DB0vVKtV9y3rp2rG5HY8RKXSW1"
    "4Dl1smmhp03hB1CpKxZrsNRQaK+NamWJbRsSmGoOYwWNxWod8PQv1gIOsPSaY/nqMHrphkN/"
    "vOrqdWbqpuNzrZ19xexaRIP1RIKmck0AgY8yBhGFtcKKrzhZg4WGhnQGlo+hFKhU7sTb737q"
    "fUAFmH7NsXxiYhIYmb34qpEHgQdXj61zOHrmupuee+H4dUkvecl5+XZilUYZK/grKiGGnNNk"
    "bRYGOsGuzfDUQQ+XEK9NFoA8EAPqnH5MztSRI0cYGho6cyl3/5eu2XP5JvuuxaXw6qGefFYj"
    "SOyKDpoqEbdQJuBwsIb5qWkaqnP8jq88cy9wGHj+Pwb4d6VSKXzfP3WpgctGHnzHTQvzles7"
    "29u2nbeunSjS4LXLk394jhNh24G7v3X4a8ATwOJ/DfAaKtz36TfuufIC78Zq3d9Zrevc7/+6"
    "+MQ3f3X0Y8BRIP5fA5wpDfT9yz4OhP/H2OC6Z6/3fwLCQncCOOfzIQAAAABJRU5ErkJggg==")

about = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAAIAAAACACAYAAADDPmHLAAAABHNCSVQICAgIfAhkiAAAIABJ"
    "REFUeJzsvfmTJdd15/e592bm2+rVXl1LV3VXo7vRIAACJEAQIiVKUFCiCDLEEUNy0JRsyzO2"
    "ZckhO2SHHf5lJkJh/QHSD7bDMx6Pggo7FKIcHsuSTYqasShIXEARO7H3it6qqmt9e2bexT/c"
    "vLdeNUACoDhAj2cS8VDdr997lS/vued8z/d8z0n4t8e/0Yd4v0/grY75+XlXr9ex1uKcQymF"
    "MQaAmzdv3pHn/K/rcUdezPn5eZdlGdZahBA45+JPpRTOufio1WoYY7h27dod+V3u9OOOvGhL"
    "S0sOwFqLlBIhRDQAAOcc1lqSJMEYQ5IkZFlGlmUYY+LrlVK8/PLLd+R3vFOOO/LiLC4uOqUU"
    "QNz9ZVmSpinOubjI48YRnlNKobUmSRIAlFIopRBC0O12mZ+fxzn3bw2jOpL3+wTe6gg7Pex6"
    "a21c/PBI0zTigmAEwVCEEGitAeLfy7Lk1KlTtFotbt26xerqqiuKIoYQgHq9zurqKk888cS/"
    "McZxRxqAtTa6/7CbAx4AkFIeCQ/GGKSURwCjcy6GiGAc7Xaber2OMQZjTAwZwaBGoxELCwus"
    "ra25siyP4I0kSbh69er/7wzjjjQAKSVAjPNh4a21ANEggqcIBjAOGJMkQWsdFx9gb2+PkydP"
    "MhwOCSEmvNdaizGGfr9PvV5nNBpFkFmWJdZaVlZWHHhPURQFxhiUUv9aA9A70gBCLB9fnPCc"
    "cy669bCwSZLEnQqHBhSO8Hye53FhA3gMXiK8bzgcMjExQafTQUpJURRkWYZSin6/T5qm0cDC"
    "+ayurrpwftZaJiYmuO+++/jjP/7jO94w7kgDCAseQsB4BhAuejiklNHNhyMsbggLIVw0Gg3y"
    "PCfPc9I0jc+HI3xumqbx/caYeA4BTI5GI5rNJqPRCGMM9XqdPM+jYQRD+eIXv+iGwyHGGLrd"
    "Lq+88grWWra2tu4Yw7gjDSDspPEFcs4d8QZCiPgaAK31kUUNz4cworWmLEsGgwFpmsbfERZX"
    "SnnE9ed5zsTEBEmSkOc5RVHEzxNCMBgMUErRarUYDofRCMLve+qppzhz5gxKKZrNJo1Gg8Fg"
    "wNWrV1leXnbh+4TPBNjY2HjPDUO+/Uve+2OcAAqP8b+Pg8BwBJcMRG8QFjYc4XljTMQQATgG"
    "bxM+yxjDxMQExpi4uwFmZ2cjKAzhKPw5hAxrLf1+H+dcXHgpJXfddVcEoeOpaziWlpbc8vLy"
    "oXt7D4470gDGiZ/bwR1wxDDGd3zYpeMXd/y93W4XrXV05WEh4TDzCMY3OzsbP3N+fp7JyUmM"
    "MRwcHFCv16NHCZ9trSXLsnguSZJw/vx56vU6jUYjnu/U1FT8HiGcjXMa7/VxRxqA1jouxHi6"
    "F1zlODgcxwbhQpZlSZZlb8IPUkqazWbkBQIYVEqRpilSyhj76/U6nU6HVqsVQ0c48jyPnxe8"
    "Rzgva23EGGVZcvPmTWZnZxmNRiilmJiYoF6vR+MO5wCHnu+9PO5IAwiuXEoZFzoYQkgLb88M"
    "xj1EeC5cYCCicyACt0ajEYFeeP/e3h61Wo2Dg4P4+sFgEDOB4NbDeQV8kGVZBIfh/Gu1Ghcu"
    "XEApxcLCAgBTU1PUajVmZmZiehleH87zvTzuSAMYR/lhl4edHOJ0+LdxpD/uEcYBYvh56tQp"
    "dnd3ybKMiYmJCBpDyDDGxJiulKLT6aC1pt1uY4whTVOWl5cjPmi1WtEIRqMR9XqdQCB1Op3I"
    "LTzzzDPxc6amppiamiLLMur1esQP4fPf6+OONYDbXf+4awyuPrj74P7Da8ddfnju2LFjdLtd"
    "arVaXMhxlk9rTZZldDodgCMEVKCMV1ZWuHHjRnxdo9GgVqsxNzcXwV84r+AJWq0Wu7u7XLly"
    "JYLG+fl5iqKIBh3OXWv9Jg7jX/VxRxpAWOBxQ7h9UcMxDqRuDwXhAqdpyvz8PIPBgKIoSJKE"
    "4XAY3zMcDmk0GgghWF1dZWtri5MnT9JutyMemJ2dZWdnJ57b8vJyRPoHBwfUajVCCTsYT/Ag"
    "xhjeeOMNnn/+edI0pdVq8bGPfYyTJ09GVjN4qXHv914cd6QBhF0UwF4wgLCgIQyMcwTjr4VD"
    "oAhw/Pjx6C12dnbodDpcvnyZfr8fkXsAhiGFS5KEqakpms0mk5OTzMzM0Ov1ovHcunULay0L"
    "CwtHspKw+OOhZGJiguFwSLfb5erVqzFd/OAHPxjBZ5ZlR9jN9+xav6e/7R0ewR2OVwPDzg4L"
    "Nn7R3ypLUEoxGo2Ym5tjYWEBIQSbm5vxc29n+KSUpGnK7u5uXLgsy9jb2+P69evs7+9HoDca"
    "jeJCdTodyrLEGENRFOR5zmAwoNFoRKYwVDOVUkcyAKUUi4uLOOdibeG9BoF3BBN44a9/0410"
    "Sqn7NOs5v/XfXePZF18+wtGP1wHG8/fgFQKzN54Ktttt1tfXcc5x/vz5+LrxauF4qfncuXOk"
    "acpTTz3F0tISq6urfO973+PYsWPs7e3FuC2E4MyZMwBcuXKF++67D4CLFy9Sr9cZDofkeU6r"
    "1YreZHp6ml6vx/z8PMYYNjY2uHXrFmVZRgbyvV58uAMMYHDhL1xj7j5QGsoh5ajL4uw/oig1"
    "UhB3ZzAGOIz74YKNp1HgjSLLMu655x6EEFy9epXhcEitVjuCJYLIZDgccvfddzM7O8tDDz3E"
    "xYsXuXXrFnfffTfNZpPp6Wk6nc4RTHLp0qUjdYj9/X2MMUeqh7VajdnZWYqioCxLpqameO65"
    "59BaR+o4UM23s5bv1fG+h4DU3mL31gt0N1+l6O+R1lpk7eW4+Lej43ECKCx8SLdCyCjLknvv"
    "vZd2u83Ozg6bm5u0Wq2oLajX60fibeAElpaWcM4xOTlJnuecP3+eNE1J05SJiYlojEKI6O5n"
    "ZmZIkoSDgwOyLOP+++9nYmIi1gX6/X7EAwEgTk5OMhqN6PV6PPbYYywtLb0vLCC8zwaw98w/"
    "c8ZN0SQndUNEsYM5uIAZbEEVp2/PAuBoehcWP3gFrTXr6+tMTk5ireXKlSs0m02KojiSHYyn"
    "bMPhkHa7zezsLK+//jplWcbqXaBvx3dswAhFUdDtdrl8+TIrKyvMzc3xjW98g+3t7RiSkiRh"
    "f3+fNE1jOrm7u8twOOTUqVPMzc3x67/+6wyHQ4B/s0CglAJnDc7hH4BDIIREcMjywdHCzu25"
    "/jifPzs7yz333EOj0eD69euRaLm9eBTAX57nLC8v8+KLL/LXf/3XPPfcc4xGI8qy5Pjx42RZ"
    "xsLCAouLi2/iGpxz7O3t0el02NzcZHt7G611DEHBgKemphiNRtRqNYQQTExMMBgMWFhYQErJ"
    "r/3ar7G2tnYkhX3P1uA9/W23/3IpcM686XmHOwIAxzOBt6oMBkNJ05Rz585RFAVFUXD9+vUj"
    "BaEQMgKYK8syLu709DSbm5tHlMXdbpfV1VXuvvtupqenI1cAxHpD8AbHjh3jIx/5SExRgyYh"
    "iFfGVUeNRoMsy8jzPBryb/7mb74vQPB9NQAhHE44v/ODi3cOgTji9sfz/dtLvWEhi6Lg1KlT"
    "1Go1ms0mFy9ejGg/SMDC30Pa1mq1aLfbbGxs0O122d3d5ebNm2itmZubY2ZmhvX19VjAWVhY"
    "iHqBiYkJyrKk2Wzy4IMPcs8997C3t8fx48eZn58HiEYcUsfwHiEEU1NTlGXJ1atXuXbtGo8+"
    "+qjHCU7ye//ocffiV37Dvfwv/+G/clbofc0CnM0QziCNRQiFtRqhFOAw1iBl8qZdH+L9uBfI"
    "soxms0m/32dvb4/9/X02Njaidi9ghuFwSJIknDp1KtLDGxsbUQza7XZjoeihhx7izJkzaK3p"
    "drs899xz1Go1kiQhTVO63S7WWkajUTQcgOnpaY4fP87XvvY1Go0GAFmWATAcDuN3sNayu7tL"
    "t9vld3/3d/mVX/kVWq0Wne4AQ4usvoB0lst/9Q9dpiSWLnnvgCJvQP0YWmg++Knf+TvHi/fV"
    "AIS1+MgvcFRAwPrFDy0LtxdJArga5wJWVlbo9Xqsra1x/vx5Dg4OovsNXiLP85jqBRJoOByS"
    "ZRkzMzPMzMwwHA7Z3d1lYWGBVqsVgaQQgk9+8pNcvnw5eosgGxNCsL29zeTkJI888gg3b97k"
    "5s2brK6ucvz4cV544QXa7XZMBdM0pd/vs7S0xNraGmma8ud//uecP3/+sHhV9uh2r2Kco5nW"
    "UI0mjVab5toSopZSExmjvMetl3/H7e1Ktq7d4ie+8Hs/lDG8vwaARTiLQyAFWOdAgBSSpNrl"
    "4yAOvOtP0zQu4F133RUXqVarRUxQliWtVisqfR588EGMMdy6dSuWb4fDIZOTk2xsbMT3DgYD"
    "tre36Xa7bG1tRRLIOcfZs2eZn5/nD//wD+PvAb+zp6am+MY3vkGn02F+fh5rLa+99lrEBKEm"
    "kGVZ1AcMBgN+8Rd/kRMnTvDMM89UQNXRXPw5plfuw+SbaLNDz+1jejnZ/nWaMwm5lNREE+QE"
    "UydPcvaRT/Gdr9/jPvrYr79rI3hfDUBRLaxKsM7hhMBav8jWOSRvBnCBcg3o/+GHH+aZZ54h"
    "yzKef/55BoNBJFpCmvjAAw/w/PPPs729HXetEIJ6vQ7AsWPHsNYyPT3NcDhkOByyubnJ9PQ0"
    "cJgxDAYDpqen+fmf/3n+6I/+KFYBATY2NmKhSQjBjRs3IuIPSuOQcgbpeZ7nGGP4+Mc/zv33"
    "3883v/lNnLXkUpJP3Q3uHJlwpIlA1aCtN9j91peZlnvIWUcxnzPauk6tu89dJ1Z/qDV43wxg"
    "98l/4mxWwznAaUAgrEIKi7tN0BniPhwyeAsLC5w4cYIPf/jDXLt2jddee42yLKnValHI0e12"
    "OXPmDN/+9rdRSnHXXXfRarVoNpu0222Wl5cjc1ev15FScnBwQKfT4dvf/jY3b97k05/+dCSD"
    "tNbs7+8zPT3NT//0T/OVr3yF48ePRw/UbDZZW1tjbW2Nq1evxqJQu92m1+vF35HnOc1mk263"
    "y9raGmfOnOGVV16J3084gzP7aO0wsk7pGghZpy4k6lsH1HYl+d5Fsu6A5N9fYfTQKySTJ36o"
    "dXjfsgBhLTiHxFHlAdUDEAJRMYEhJQu7OUi2Wq0Wn//851ldXWVtbY3RaBQvbjCELMs4f/48"
    "SZIwPz/Pfffdx0MPPcS5c+eYnZ2l1+vFIky3241K4A984AN84QtfYGJigi996UtcvHiRWq0G"
    "HCL7e++9l5MnT8aMYnFxkbNnz3Lq1CkmJydZWlqKhZ/gQRqNBtZaZmZm0FpHEDg9Pc3S0pKv"
    "c+AQuUCZKYRtYY2gKEeMih6Dfg9jRxw0DXnboXoaoy2psTTskJf+9n9/11nD++YBBAawWAuM"
    "cfmxymcdTh5mAONyrzRNeeyxxzh58iTdbjemi+G9Ib8OXcNKKT74wQ8yOTnJ9vZ2XMyiKKIE"
    "LLjksKDNZpNPfepTPPvss/zZn/0ZN27c4LHHHkMIESngz3zmM3z5y1+mVqtFl3/jxg2mpqb4"
    "zGc+w1e/+lVGoxHD4TCKSNvtdhSUhBQ1hDhfElbs732D/S1wYppaY556No9U05jOEHeQIUaK"
    "xGTkakQ60UBboH8dpaff9Tq8fwYgEoSzSOconfSg3zlARvAHR+cBBAnXww8/zOOPP06WZRRF"
    "wWuvvRbZtdAdHECd1ppHH300ho5xFW8o3gTvMU4xhzj94IMPMhqNePrpp7nrrrtYXV2lLEsO"
    "Dg44ffo0Z86c4fLly0gp6Xa7rKyssLi4yP7+Pjdu3ODRRx/l1VdfjVRvWZYMh0OazSbXr1+P"
    "YpRWq0WapuSFoejeIN/4OqaUDFWKkhkynSFLYXa5x2DQprZrcJSIyRqlEMjyFpMTK+96Hd63"
    "ECCThNKWWJEirPEPV4LLcdagK9ImLPx4NvDQQw8xOzuL1ppLly5x5cqVqCQOxZ5er4cQgo9/"
    "/OOxEBOaQkKJNmQL9Xo96vnGiaVOp4MxhpMnT7KwsMDLL79Ms9mM7WGDwYBHHnmEPM+j4Rlj"
    "ePHFF/nWt77F6dOnWVxc5PTp0zjnIugMFPTi4iI3b948YnhKScqiRmamyNwUNVNDaQuDLdzo"
    "FuWjTXg8Rf3KEsl/8yHsgkA4TT7q0GwNeeb//R/fVRh4/zBAxQBiK4avKggIR4UJDhnAUC4t"
    "ioLV1VVu3rzJpUuXSNOUr3zlK1HnVxQFx48fR2sdY3movIFv+To4OIiudzgcUhQFaZoyGAyO"
    "9BO0222cc1y/fp3t7W3KsuTatWsxZJRlSb/fZ319nXq9jlKK4XDIK6+8whtvvME999xDs9nk"
    "y1/+Ms8880xc9OCBNjY2eP75548Uo6LoVUpKV2AoMFgcEicSMBkjaykoKBuWcjnDCoMETFnS"
    "dF1OLAte+db/4J78yu+8I0N430KAswaBxFoD0mEJUeAwA3BOHkkBsyzj85//PI899hgXL17k"
    "4sWLXLp0KcbUqakptre3AThz5gxra2uRtCnLkosXL6KUolarxdz9+vXrPPHEE6yvr9NoNJiY"
    "mKDf79PpdKIEbHNzM3qgl156iZ/5mZ85wkaur6/z/PPPU6vVaLVazM/P0263eemll6KuMBhX"
    "qF+E33X69OlIdvlehhHWVmJRkjH84zBC+KKZPWyEDUeqBJ2bF6m35rh7cQVxYpXOtf/LPf30"
    "0zz2ud/+vvzA+1cLqDIAYY3nAMYeuMMW8IAHer0e9913H5/97GejqvfChQtMT09HEihNU/b2"
    "9jh27Bhnz5490mK2u7sbd9jW1hbb29uRBgbY3NzEOce1a9c4deoU99xzD/fccw+Li4usra1F"
    "Dn9raytyDGFIxdmzZ6N28KGHHuLxxx+nVqsxGAzodru0Wq3IXwQqOPAQIb1M09QLVgBjLM4e"
    "St3Hq59w6CHHD4lGpZuUxXn2b32T3Wv/D2Vngx//6V/kiX/x29/XG7xvHkC7glQ3QRikVSAc"
    "QgIWwAO0ei2LWv0sy/joRz9KvV7n4OCAyclJ/uRP/oSTJ09y/vz5WNiRUrK+vu559U4nEjWj"
    "0Yjt7W3a7TYnTpxgNBoxPz9Pt9tldnaWq1evcnBwwGAwoNPpcOXKFfr9PseOHeOhhx4iz3Nu"
    "3LjB7u5u1A+GEm+z2Yyg8rXXXuOVV15hbm6Oz33uc/R6PZ5++mnAM4bT09ORB7DWUpZlFIwE"
    "4AoC6ytkhOqwVGCdBic9QeY8i4qTVSk9wZk5VGKBEpyl7D6J6J8jkWe/7zq8fx5AWZwGJ9zh"
    "l7H+pxCCpAJVgftXSvHRj36UsixpNBo888wz7O/vx50fuP+JiQmOHTt2hDyq1WpsbGzgnGN5"
    "eZnBYMDk5GSMxwsLC0xOTtLtdvnQhz7E2toaP/VTP8WHP/xh8jznu9/9LmmasrS0RLfbZW9v"
    "L2Ynxhjm5+fJ85zhcMjBwQErKyu02+0IDm/cuHFEEBp2tVIqik2VUpFCdkJgrPWlEeuw1uGs"
    "xTrvCaTz1ZMKQFUPi0xGWAoEDkECTuPKA7JMf99leF88QOfpLzlr98AahAMbXZrAuaqXTwhc"
    "tUMAVldXOXfuXKzBh+JJoH7DxVxfX2dubo5utxtLxsPhkJmZGW7cuMFgMGBqagoggrKyLOn1"
    "eqysrHDixAl2dnao1+ssLy+ztrZGvV5nbW2NL33pS9y4cSN6kqAzDDX+kB2EquN3vvMdBoMB"
    "p06dYn19nW984xtRC7C9vY2UkosXL8Zew4mJCU+NV0YiKnLMIZBKHC62EIjvN9+rEtZ4zyHQ"
    "ReE96/c53hcPYIoShQFnwTosBuesfxzCwQiAiqLg3nvvpdlsYoxhMBjw5JNP0m632d/fBzzt"
    "K6Vkbm4uTgYLYGtzczPq/EIFMGjzQj9gr9fj3Llz7O7uxpbw4CG01rzxxht84QtfiHF/fObA"
    "uPx7NBpx5coVXnzxRc6ePUuSJNx7770cO3aMVqt1eA2qVrQgFwPGBmFVGMDZCh95etzDAM+S"
    "vvX6Hw7LqoqsmFLz8Md/9c4CgYYSXIYQfvFxFrA4Z3DOeMt3xAzAWhubKAaDAb1ej5s3b5Jl"
    "GQcHB5EgWlhYYGFhIRZlgpsO0q6ZmRlu3boV429RFMzNzfH666/ziU98Ii6ulJLJycnoGca7"
    "kc+ePUun02E4HB6ZJBIEH6FEvb+/T7/f59y5c/zVX/0Vf/AHfxAnh4zPFLh8+XKsdNZqNQRQ"
    "OYFD9++E/4nHBEJU2EC4asNUnsIdqqac0yhbo3wLxdX48f5gAFEitQJhKh2AHXsYjwfE4Y4Q"
    "QvDAAw9EqddLL70Ud+bu7m40gomJCWq12ptkWVJKZmZmmJ6epiiKCN5Co8e5c+di06dzvvs3"
    "kETjjR2NRiO+Ljwf0Hy9Xo9FqIWFBR5//HF+/Md/nJs3bx4BfWGOQKvVit+t2WwC0Gw2kYH5"
    "tGCNI4B/byNBEWURcqx+UhlC9B7OAgZFE/s2GsOkc/NbLq21SLMmVljoPMvBQYfBYEiaZazc"
    "/5/+yFWK0oGzpbdi68Cpwzhlg8zboirdfavV4vjx4/R6PfI858knn4zUbr1eZ3JyMrr7RqMR"
    "pdjjwxpCX2Caply7do2TJ09Sq9XY3Nzk1KlTEYyFmB5CS3DPRVHQ7/dpNpuxqhd6Azc3N8nz"
    "nIODg1jr39/f5+mnn8Y5R6fTYWJiIvYOjnc6hToEELuNg/vHgbDVaDwF437/UEEX6iVjsroq"
    "BMhU0R8Nf+BaJKp+gXrjAQozQKY1Bos/z+SSYJYSp0fovUdcv7PBQfcqhd2GpMGZe/+rv5NR"
    "CAzOBk0gCKtw7rADWIjD3kDnHHNzc6RpGuXWzz33HMeOHWN7e5ssy6L0atz9Bxl4o9GIrF/Y"
    "tVevXuXEiRP0er2444NxNRoNtNYURcHMzAzNZjM2eszPz8c2cGNMdPu3bt2KlcigGbh8+XJU"
    "En3oQx/i/PnznD9/nrIs0VrTbDZjzSIsnFIKU3kvZyvPKCXWurcEcuOFskCgVf+AcwIha4xG"
    "3z8DAEjS/BbWXiVJhhidMsH3cGRYMYlQdUxjkcn2IpPiISjADG7SvfI11xlsMMh3sEmGdIaz"
    "H/wv3rFRKCt8TiuqdMYppJBUnh9rTUwF8zznrrvuijv+xo0bbG5usry8zOXLl2MP4Orq6pHx"
    "K+MAMlQIgSOfs7i4SJIkjEajIxKyUBMIqH68n3B7e5v19XX6/X7M6UPKGT6j2WxGUakxhitX"
    "rnD9+vVICAUACbC1tRU/a3p6mjTLqlTRozhhHVJ5YBywnxB4Eg0RKXTGyCGBV1tLVaPQP1hl"
    "nJRmFtns4cQUzgisNOAEwvVBd8EcUOSKRDZApIhGjYn2CSbEXQDYQZ+8v8fmq/+b6w1uUVpH"
    "rXWMU/f8yvdHnnmBTCWl9bE0FRJnJQKFFApbAZeAxu++++64g1988UXq9TpbW1txoUajUaRS"
    "AycfOnDDogYFUZqm5HnOzs5O7AAenyscpF7BXff7/Ziz9/v9OB0kpJhaa65fvw4cNpzs7OxE"
    "ALqxsUGr1YqK4t3d3VjSDqCz0+kwPT0dm02cE5SGCvBZpBCkSKRzKEAikIAYE8uKyuiDHVhK"
    "pJqk0zn4wQbQOv6rYrj/P7nMrGLSTXD1kIQhUAhyQKJtAU6CKnClBFcHWijpyKZbLMx+mAUy"
    "0EOK3hZbr/+vbjDYo9B9tIN6usBdD/5HVUDrYamT2BoJNbTIwTiETNBaVDUBH9e01pw+fZrt"
    "7W2MMbz++utkWcaVK1cik+eco9VqkVW7J7jU0AwS+P8w76csS/b392O6ODc3F6ne8L6w4KHN"
    "KxBFQfIVmj9Ho1EkhkLY6fV6sUEkeKrPfe5zvPDCC2xtbcUMJVDXvV4veiePOfyuts4iZYjz"
    "t++nSkw7ThEf+TOopIa2/R9sAAC2mAK1CTIDIyLW8JyDDz4CA6IEXQMJUo1AdtBaQSGx0uGs"
    "xooM0W6zMPMBsHWcNph8j4POBpee+z3X797i8tYWWVanUWtQUzVqtQyVJt71Sxn7BGTlVpeW"
    "luj1eqRpGgsz7XY7lneFEOzs7BwZ+hAWN4xwCeCsKAq01nQ6HUajEZcuXWJiYoJ2ux0vSsgA"
    "JiYmYrNGo9HgO9/5DktLS9GwQt4fPBDVLgy1guFwyEc+8hHa7TZf/epXabfbkbkMADKoj4EI"
    "QI0xHgTi4z8CnLsNBLhDDHA7+KtqR2inSJLa2xtA69gXhen8z86UayRVCPCL7nBG+jgtvFgD"
    "WXrr1CBIUMLiKHDGIZEol8NoSOlKbzxK4dIaM8dmmVtcgVJj+lfoHxzQPdil0+9QbhSIxNO2"
    "k606pjQYY7HWMD09zezsLLu7u+zv78eYPj56rd/vo5Ti6tWrsdEy0LTB7QftfyCDQrpojOF7"
    "3/sejzzySHxd+NzA8AXP8+yzz/JLv/RLkXl0zk8HKYoizgNwzsWO4vvvvx+AV199lU6nE+Xj"
    "QdkcytJhAplvWvXcn3P4a16JZLxr98YgxFvXdkT17wBSCEaF5uOf/q9/IDaLVPBgUKPR0BUQ"
    "s8HIEBVR6+v0lWlxaGWevavSEKx/rzRIqpYurZFSo4s90sQyGvYxWtCcm6Y9t4gTCmc0g84B"
    "vcEBnc4eH3pgkt3hOrf2ekxOL5ClCf2+1+z3+3263S5ZljE1NRXn9tXrdW7cuBGzAyFE1PWH"
    "Iou1lsnJSTY3N2N3bwB5Tz31FPfeey9zc3PU63UGg8GR6WFPPPEEU1NT0SjyPCfLMl544YUj"
    "qqJALbfbbV5//XUeeeQRtra2YsoXwksYIhGMFALvUe1eL49GSYm0PveXwpElIIVD4Dcc1qun"
    "nfAcSjQGKcjfAdMfX9Fe+g9EsfdHToqJ20qPYz8P19+Hh0PTqyxWxC9wJC+tGj5dRZyopMso"
    "P/DPI0lsg3ojpTWxzPLqKfbtPHffN2DYH9CsZbjhC6QyZzA4wBgX6dydnZ3I2oWpW3/zN38T"
    "xRXnzp1Da83e3h7tdpuyLOl0OhwcHMSW87ALy7LkueeeQ0rJ7OwsUkrPi9auAAAgAElEQVTO"
    "nj3LcDjk5ZdfZjAY8PDDD5NlGYPBgJmZGUajEVtbW1ENlGUZZVmysrLCzs4OtVqNbrfLL//y"
    "L/Pd736XjY0N8jyPnif8/oMDD9Q8aXU4FUVUqXAiBR76VZtShgtehQEpwIkjZWKJRJu3nzp2"
    "xERsOUBmzQpNep2+cGMLX7mCo/Xo6gVjjiYi05CfCkBUyF4blHOk0lu1dQVO9dFIihKUlqjB"
    "DtiE+kSdUx84y9zMWZYWh6xMXeD6+UX+5GsHsUQcBjBevnw5hoX9/X2EEDz55JMAMe9OkoTd"
    "3d2I4ANwC7E3iEXfeOMNTp8+zYULFwAiHgDo9/u0222UUly/fj1SyiGfl1LS6XR48MEHabfb"
    "bG1txfMKmcn41LOQ6QDVrKDaYSOstagQ48eLQNWfw+dY6xDCxgIQziGUojTvwgP4L3qASmf8"
    "09EIwqoeLviRGFTVrcePQEo4F19Q5abaf6apV4ogjZQKXWYgcqRyIDS5KnBCo4s+styiv79L"
    "YQSL6yf4uce/yPKJJ/jH/8tfIoQ+Ul27detWBGFBYxfy+lu3bkXhxeTkJAcHBzEWT0xM0Gg0"
    "Ik2stY70bcjxAyG0urpKnud0Oh1ef/31I2ko+F3c7Xb55je/ydLSUlzkoEYOhhKAZqCqgcog"
    "87j4TggPiuPih9SwypREWA98luDACVvxBQLzgzmgNxtA+/hvCbP5T51O5zCyRsIArDyy3293"
    "7bjSNzJUREXs8atOAmGxxpC4FKNH4HKUrGrZVR0gTUexDCyEAKFxto7AkKSWpOxRlgcM81tk"
    "OH7swx9h6+c7/MsnLnCwv8tgMIjAMLCCcDh6Nejxg9sPfQOBmAkkTrPZRErJ4uIiy8vLXL16"
    "NfbxDQYD1qt5Q7VajTfeeINOpxOp3JA2jvcwhKwkTVPuu+8+XnvttSM6haA8DiXvWCRCoA3I"
    "pJqGYjVCSJQAJQVSCC+kQUTP7BwolWEYgExJ0zn2u9tvawBvIhhHRiGSDsrmCJt4YCfGGzcs"
    "iGrxROXeKyn3kTp1SCWr/zuMF39QiRsqgYN1Fmsjx+X/MwpBSSI1tVThpMW6jNFgyCi/zGh0"
    "kZ/7yY9x/wcWGOWe+g20bUDXcEjMhNJuAGhhykee53Go09bWVhSVHhwc8O1vf5tOp8PKygrX"
    "rl2Lny+lZHd3l/PnzzMxMUG32wV8mJmZmYlFpkAYffazn2VtbY2f/MmfBIhj4cJouPH5RkFH"
    "4KAq//q6AMG1B5aTsMdExGY+pHhD8gxy4nmbtzneFCRaK39f9Lf/qWsmltJahKp2vAhQ4Ci1"
    "KFyFCaKbCLmCiH/3//kKlcWhglGJw1BxWMasMk5yjM2pZYp+7nAGnDE43WdYXKfZUPzCpx9i"
    "NCq5ePmAJFFHhkwH5VAgZ8Li3rhxI456CYLM/f19Go0GGxsbsSE0CDbCxK+ZmZm40N1ul5s3"
    "b8aScUDxo9GIdrvN0tISZ8+e5dy5c+zv7zMcDvn93/99gDhIOhjpYDCIHsCzjt6XWmfBipDa"
    "e/q3MgD/3sPY7COEw1ECFmsgkTVM+dbp4g80AABbejZKpQbtDi3vyMeNkUVvflpEo41eg1Cw"
    "8LveS4Hs4XvGy5ZW4kyBkCW1NKEvUrQ5YDgcgG6C6wFbzLTP8u987kP88Z+e55VXL5JlaSzc"
    "AJHnDylbWLTl5WVarRZKKf72b/82EjfBK4TXTU5OxtStVqvFNrILFy7EwVCBDwjMnrWWy5cv"
    "c/HiRWZmZtjb2+P555/HWht/Z+hMDjMPQ12gKAry0Qhf2z+q/BWV2OPwOo2l4/HwnIF1IGXK"
    "YND7vgsfjrc0gPbyr4rB1j9zNZcgnELcRkMewYCOirUKwETjhI5MFcKhnABnESZH2sLHHeu/"
    "lhMWi0HYGhI8kWRvglVImYJTWFtSGuj2e9ikJLFNdKmwbpvF+Xk+/6mSf65HvHLxJok6nBYa"
    "MoT19XVmZ2cj/z4ajZiZmeHZZ5+N2CCwhmGh9/b2mJub4+LFi3HAQ57n7O/vc+nSJVqtVszl"
    "u91uLEqtra1x9913c+nSJTY3N48oiIMhBowQMoJxD4DwmZJzCU5LnNQkTqKkQ0lLkoASFpyo"
    "5ij4Q6IwCUgtSUSH7do0ShyOuH9XBgBQ0KfulhGuOGJmRzXpAjAcRvrA44/Xra0vTR6az5GU"
    "EVfFfedTQkmOcxIhwFrNzs4tirIkzwt06RAoVDIFQmBMTqm3WV25h1/6e3W+9vWMv33mDYQA"
    "pXwOffr0aY4dO8b+/j6dTofd3V329vYiCRTc8fiChPx8Z2cH5xynT5+OUrFXX301GkkwpuAF"
    "Qt6/tbXFmTNnuH79Og888AAPPvggg8GAmzdvRvVwSB0HgwF7e3uAnzqaJEmV4mkQHv+AHUP/"
    "R/P9ECL8klgQCkhx2pFm7zINHD+mj/3nYrT1B06l2dFaMwFwhAUOLj6Ak0qhEpKBKgw4F060"
    "cm9j9oCzCGFxpkS7HjiwWASSF1/6HpOTU6RpHSESkiTDiRSjS2r1FGc1I/caiwvL/Lt/7zHO"
    "rD/Lt5+6wtXr+xgL165d5/KVK+RVKAggbHFxMS5wwAyBxg27MhSDAj+wu7t7pJoXJOvBswTR"
    "Sa/XY3t7O5aDb926RZZl9Pv9yPuHsjEQBSGeD6iyCGNRVTYu1dE7ighCxjR2EavwKpzEodCF"
    "4GOPv32J/m1MpAN2Diec36kiLOC4NRxSgpE1HGOlIocRV5vb0Ky3DD8hpESbHOuM9xrC0Wo2"
    "6PV71GsOKRJSVWdgc2SicNahSMhFwTB/gzRp8xMf/TAf/8ij7Owe8OyLl3jh5ctcu7FddQrX"
    "6A98eXdxcREgsonjQyeCMKQoCur1Ov1+n6mpqXjLmfH7BobZREHyHYQi09PTrK+v85WvfIXl"
    "5WWuX78eVUvBgELPwnga6KxXAmnj9RJSSi8RRyCEAlk10kY94OG1d1iskEiXYH+wFPCdGUD9"
    "2G+KfOefOIoF0nqXUZGTqoYXc1QrGDX946cy9ncnBBaDcjl6NEDhADPGK0mkS7HsYPSBZ76M"
    "QAqvEjLW4UgZ5RaZ1MhLS60SVsokwTlDhsA6idM9BsU+Mq0zO5fx2I8f41M/tc7IpLxxfYeX"
    "X73G5TducXNjjxdeeBEHZGm44ZMiz0fx/gOhYyiIN8PswIAvxoFbAHUTE55Gn5qaQilFu93m"
    "2LFj3HXXXbGaGPBGyFKCKhmqCedKUdqcohghbQLCIktNkg8QoxqiX8cmPsQJUY/sn5ASZWok"
    "6T4HyTyFrv/dDQCgMH0atSZGJySqUu2+ldDwSIoyfoR2r7GwceR9FuE0RpcV7Rw08D6eFaUl"
    "TRqUhe8hPMJ6VTHG58qBbZRgSoyzjEYlhezinOH4nGLt4wvUfvo41rXo9i2vXrzJ0y9c5dqN"
    "PbZ3DFK26PX7NOp1ZNXIAd4193o99vf3Y6rpv7J3xaFNPc9zVlZWWFpa4pOf/CQXLlzg4OCA"
    "P/3TP42TS4K3CazkYDDgxAk/3UNKXy9xxoc/HzJ9HQXhNUDWVRSaNUDvMCw4hRQFthiQtKYo"
    "Kk7g7Y63NYCEKYzNEc5bq3GGcX5YjFeHqp9HjSCkkCF0HK0leIgwxOgc5ywpEofGusSzXzKl"
    "lrUwZlhZjs9zD9UvvhgipdfTKSnBGnRZkiqLMSOEKHHaYYChlmjnUDLjgXMtHrr3QxRO0e0X"
    "7O7nXLi0ycZWj9dev063P6LIc4ajnLKK7SE9vHjxYpz1F3iC8Qmgm5ubnD9/ngsXLjAzMxMN"
    "JdQgxusQ41oCV11DawxWVllWxQ14mRve+KMc3F9FZw24EseIRCmK4Tvggd+JATSO/QPRufGP"
    "XaNeYHUC4mhWMMb8Hll0MYYLDokLUcWymMV6IsmMgBIlwGlTeQyLs5AmNVoT0xgDeT7wFweL"
    "lIfnYEyQRPtfI0ziX6dLv/gWpEg8P26dH05lhhRmQClvYbWh5hQrUxknHplBpSto8UEEdfoD"
    "OH9pm+e+9zovv6bJS2hPTjEx0WY08kYbduHx48dRSvGXf/mX/MZv/AanTp3ixRdfjC1iExMT"
    "USkExOdDquhvV+PQZYm1CmNF1SNgMZV8zliDiN9fRsPBOQ8aTUpKDT16Z4r/d9QaViszyAbY"
    "1CDzCYQYty5fiYp/c57hi1kADiUdthjixACFwTiJJANbYGwfZ7uAxDg/Lca5EpIWQjaRsk4t"
    "S6jVU4YjvDu0JQKFc8a7/sp1WmeQVmDcyC8MAkhwznrX6YwXnzqHdAkSgSkMEuWBqLSkKqff"
    "2SVJayAFk6niIx/IeODcR5DyUYa54urNDtsfuYfnnn+Zp597kTzXDAZ+NsDp06f59Kc/zZNP"
    "PsmJEyf4hV/4Bb7+9a/HKWLjt5UpioLp6ekjYlQhBBhBmTuwGqcShHEYbTFaU4wUVoFQDil8"
    "kSjUBigUVg5RYp5OsfEjNICTf18MNv57h26gkhJjJFRFf3F7VsC4R6g06/6PY2mhB4/Wajx/"
    "HbIC/wJJirMCoQT9fo/l48u0XbsCS9L/XiFi3EdYX2cwDuO0xw8OjPMDIROV+JG0zlHoAiUk"
    "zlRsYZqgS119niQflSRpnaIibUqjMXaEdTtIlTE5McPD9zWpZSt8/mfPsbn9c5y/vM13v/cy"
    "b1zZ5NKFV7lw4SKf+MQn6PZ6/Iu/+Is4Hmb8ngdZlrG7u8uJEyeo1Wrs7e2xubmJUgkIr4y2"
    "VvqdXxFHWnsOwbPqXi8Q1M9CWrRxOKVpOEmv8/Yk0Ds2AABt5mimJcb1EGRvXf2L+rTAA4RU"
    "8LB9CQfKAU4jXI5z+TiSwIePGkIkng9w/svXG3W/eytzsc5ST+qUWoMJvYX6MAMRoGR1FxEp"
    "kEJiK3dtSl+WTtKksmN/3qM8p1EBv0atRlGlZVJKlIVECoruAXrQJ5nMcVnGVFvwwQ8YHj7z"
    "cVRjgoFO2Nrt8txLGzz11LcYDvtMTc3S7/di5hC4kNFoxPr6ehwhH9JBXygbx1JVp5DFG4Ez"
    "vk9AVHdIEZUiyCY4aRkMDVJlP1oDmDz+RZHf+F0nkslYjvL5veAoUyzGeIAgWZaEokUAANI5"
    "tCiwdoQUGbEpAIugjhOyKneGZlFRKW+Ml4274Akk2pU4Y6vR854mVZXGMOTbSZJ4I8C3Zjnj"
    "DaMs/EXX2lBv+Epev9dHIkmVn2PkrEOSkI8KL4xWhv6wS4MmrXqDOo6Ofh3RT+l0hjRlxo89"
    "OMnP/uyvc+n1G3z5//jnvHq+OHJjqCAFO3bsWCwlBybQHcbPeE2jGNyJqnPIRU9ohfBwwNXQ"
    "ZYdeb0B7cu0dreu76g0cMYEUA6zNkUIjhca5EmH9A1fiXIlzGmENOIOsOACjc6TwluyXoYPV"
    "Q6TLcDb3JSOXIZI2TvjmEedcJXAwOKtJU4VzJVLVqkUBYTXS+cWXSKRUPiZaS5amSCAR0vfU"
    "O4fCQ2shBbr0M4nLsiRLMpr1JsP+ECkkCEepC8IMQ4tDSenDhzaY3GJyy6Cfo0uHHTl0f0TN"
    "lCRFj9pwk86FP2N9qcU/+NUvsHZ8AWMOb3GjtWZqaoq1tTUmJyfjTSTKssA56bkQJ1AV8lfV"
    "yDxtHVooPzMIiW8cBWccmC7OJfSKGo9+9rfelgV81wYwtfKfiNJIpAh32vSu/bCufziwwEWG"
    "cDwMgK8GGpzTVWroPYVwgec+rHFLARUJjsNRq2WEZlJnDYkSGFvGtNQ4M9YVDM5YrHVIcXgD"
    "amctUihEnEcsabcnmZmZRpfeRQshx75LuKdfdR7VKNsiL8jzIcZoRvmAiVaLLM2o1eu0J6fQ"
    "usAZy97uU6wszfHoR+47QirV6/V4L6KiKOI9iJIk9frJ6Pk4knWFdnF/uPjDy8ISrFO3Fex/"
    "8PGuB0RonaASG6tn4QipUIxb4yXreFQtzs5gTOmXuxKUOKEQIsFVya/Aa92SNEFJSVn10Qnh"
    "R6n6+gE4o31qJARKHr3DSKkrCRqArTSKeNdfSzOsczSqAstwOKq0A9L354VUNcRYa6nV62hj"
    "kJVWcjQqCFo/o73BFGWBQKBtQSNrk7NLb+MlHvvJh3n2hfPsdwZx/lHwAPPz8zzzzDNAdQNq"
    "Y45cPFGFungVXdX0UukzvX7AYZxAyBTHO2MB/ZV/l0ct9S3dxvgpX+Pqm6NVwkMNwVGNoEDj"
    "fByXtsISXh2sZBIzhWBcMpERaDZbrSrl8XRymkiajTQqbW/v7Ameqqj69xyONMtoNhtoYytB"
    "ZsFolFfsnKHUJUKK6Lw8whaRvvWkTbh1rWA0yun3B37699APqux0uxitGPb3cP0m5eB1Jlo1"
    "fuKj9/nvr72W8f77748NKS+//PJhw0kcCOHiNTxSQI3imcrbVilvXjiSLKMs37kPeNcGkC3+"
    "Z0KXBVImlLaPsnGZwZOVyLGHcg5hBc74vn/nRmD2Kjfu/by1NXCJby5xDkrjq4HCkQiJLb3Q"
    "URc5zUYdXY6wVmOMZqI9zdzsLImUnj3TGiV9Xm+MRUrlXXNWw2qPExSKiUaTVCQ4YxkOBpRF"
    "gdWGRCqvVTHVDSkRERAmUqGL0s/l0yb28BttKUc57WYbnWuEcTSyGknaJE0VQgsGOy/ygfvP"
    "0KgngIiTQYEoMwtdQQCmypgSIagJn/N7HYVDWP9v1oSHJ4ISMUCIRfrunaWAP5QBAKjkGLnd"
    "RLk2CBPCdAxVh669eq6KpQILRntwKHzcN2aMJYzoN4xzdRitsUZjSj/5qlZLo0cRQrK3t+8B"
    "YUXwBNl0kFYbo2nUfb+j0YZBf0Dn4ID93X16vR6j4QhZtaOrpLofofPkkr91jfKzinSJNtqn"
    "oNZWdzOxJKpqLBWSzkEHo7Uf8mCdb3TFZyyjXofFhYwHP3CKsiziLeaCJC2wiP47ePwUVFXe"
    "G7nDB4T4RPS0FoQwIGrod1oK/GENoL78y8IahUJhxOEsu6hnhzH35LBWe+7aOV9JtKFw4/xk"
    "sEDsxC9a8QuVOkZXcd4YTVkWGGto1BsgPJETGjZDc2hZavLRyFO/xtLt9egPBkjlB1Pmee5d"
    "vTi8FW04nHNkoUETH+ullBUOcIfDHqvVkVJ4PqEKFUVZxu/trPXt3UiELig713n0kftQCubm"
    "5lhcXIyCkCAQFcIXfAIG9CHvcM3H3f/4dbfWoo0ClaDfpiX872wAAE13EsQVjEsrOxRjQDCu"
    "oz/twPqF6d948CbcodIlVPfGiaSwILYae1Lq0u86mTAcDWNff1EUlNVQB6kkWS2j0WwCfjGF"
    "OFxApRRq7NbwAWuECxoEnkJ4VZKQIn4fqQ5vW1dWEnNjDc76ErCAyFWYivMPpE4qS0bbNzmx"
    "vsC5u45zbHGJjY0N+v0+165dixVCpXwHkF/3wI0enuuR+B8P/6rCWITM6HU773gdf2gDSNf+"
    "PTEwKanSOAFOVDJmZxBUD+eJGeOsvy+AzsGMKm5eADWk9KoYWy28dRaDJ4okDm01ZTHAmMKP"
    "lRP+ciipGA1H9Lq9eBfwLMtAKBLlB05ktRpSKtJUHRkadQgSq8nkSpJIhRLCYwBAWFthTT/H"
    "WDiBMM4LmS2eCyhKJJKk4vGTNPVcgQrDHD1elUgkOcJq9MEWn/ixe5mZnmQwGMTGkNAqVpa6"
    "Ar0CIT37GIwhkmkcDpDyj8pjmBIraiS1dz42/u80JErIZbQ+QKDBeVLIT/sae+A8SSMc2BxR"
    "dQd5v6ZAJNUOs3H3e/2LJ26kcJTV7ih16b2HdSghKYuSMi84fnyVpaUVz7RVF8c5i6lQfVGU"
    "sZXbVGmc1jqKKm3ViZwmqU9BnQd+Pr+uppabwwKXxyUGJSVWG6baU1hj6XW6lEWJNX6oA1WK"
    "iLNQJgg0w4Mb3LW2hHI5+/v+7iRBkhbnD4caf1XoIvRluJACgqfMZXzgvI5iOCqZnVt+bwxg"
    "8vh/LLRNwehq3LsF60Uj4YHTvqplcqw1gDmS04zn7R4/uJi7e+zoiRNdln5wlFLeLeNlU/V6"
    "3cvBi5xGoxnvCGasL6GGeruUknq9TjKGvOOkj8ob+BMhfn7ksapw5irDktXcHq09LhkOB4zy"
    "UeQOHIc3ucLaCssolDCUo1s00oL1lUl2drZJ05SdnR12d3erLuYwKDOEoarahzji+t+EAZxD"
    "kFIayad+8Qe3hI8ff+dJocomOPKIVh3yCI3t7wzmkbwwRQVXj9rdkZgW6x+HObDWOvbI2cp7"
    "OOuniOVFwYULF1DKo/I0TcmrXr2gvatnGe0Jf9+ebrdbGUE1Gh6BNT6TMcagRLghdTXCzlJV"
    "3Sq2TQq/w2XFRzjh7/4RSStPkBnrc/m08kpe9Ao4Q9m/zqlTH+T1qx2GwxFp2o/pn6hIJyHU"
    "YaWvUoG4kG1XfRaHyix/HlZKyneeAAA/gjmBrZP/rdC6RFqDQyHsEDgsyVoLuNLr/Z2ffU81"
    "DygUlQhpo/PjUbVx3nvgE10hRVUerUanETE6RmtCQ5qSMuIDZx1WW5SQWAvdTjfyAuDRdUii"
    "ms1GrCGkSUqaJCQqwWrvRawx8VZ2pjyc4pUIBc67+yzz4cNZg5QCa3R1AygfYhCe95BWQNlj"
    "uuWYbDiG1dyg0EfoR8SGIpithkH5L+01DWCdxDovbvEP4XGWmsC+KyL4RzUoUs6hVI7Vhc9F"
    "XY6gROBlStjC9xdYE6I7QdDjKlrXU7verZ1eb7G6qJhsGaSw5KMRuiwwpa5m38ojVG2SeAWQ"
    "0cbTxyrFaL8DdWnIKn7dVESQNd4wWo0mSkp0UcYMwWMPqNUyWs0GjVqdLE1iOqgqQ/FzGi3N"
    "eqOacOY/0zk/njVkHp6yrSp4CJyx1IRieLDJvWeXMcaLSbTWVRbgq5aeKwnppsdHFjBOVAsu"
    "MJbDBxZL/Ujd4J0cP5Jh0e1TvyU6r/+XrpbllKYBaDyXlXuTNQXS6MoYFEgVMwbPbxmItLJj"
    "KpM05gqMsZTG0S9hMLT0hlBqv/uE8zs+z3OEVBjryLLqxpGVvNsYQ6PZ8MwgPkUb5TlK+Pw9"
    "lGfTNGVUTSFT1bmhQSUJUlq0PpwUYp1FImlNeAFpqYeVI3NoPMC0zpEqf88jn7r5qp61Gqz/"
    "3N7eJidPzPO9i46Nrd5YK3uBdR7YySr03AaaqpL60TTQWsiSjFyPP//2x49sVKyQ8whb4mQD"
    "JRv4li6HNSVOF1hbAtrHQufvGBZyKlfx2b46aEmlJlNQS6FVE0zXDcvTOaeXcs4u56wuSKYm"
    "QClPwkxWN3jURnv3awNRYshHOY26Lx9PtCaoZVlF6NgKJ0jKokAKSZqk3rvg72wewB945F8U"
    "BUr6Hn4hBJPtts86DkkLtDEkaoxRrCqZ1lowjjRNSGRCI7EMujc5udjmwsVLY3dF8ed6qPYJ"
    "S3SY/78lEeQsQtXQ79IAfmTj4pvZAnl+g1SkZLKGTScZ6BJt97HcRNmCBIOqqgRCOHC+IuhI"
    "As2FwKIcvpxMFetTgdf2ebVslsBcyyKcoTs0uLTAFHU6vRFRfCQBBHnhPYmQImYagUsoi5JG"
    "vcFwMMQ4XU3urFd9zMYD18NYE1O/Rr0Rp5PHvVkBtNSLGqv+ycOpSrKqTran/NyiREp0vs99"
    "Z87SbHhQOhwOsQ6ESKuZgN5TjUGlyBSKKkuIv95ahEgZFO9MDRyOH5kBqLVfE/vnf9tlXGFz"
    "X2M0NBp1atSpiw+g6z1Ke0BZ9sDlKGoIkWGNRciiQrSmsmZRlTp9XSGrmkQcDqccCWCF752f"
    "bVi03WVyTiEXErYPRoysojOQJIkiUCi6KKhlGeGmJKGLJx/lVTeO83IwIdCFR+xegFoRQlXW"
    "YbRGVR3DfpybF4rgIFGqyh4gEaEaaVEq3IzKexlrDMpl1K2lKTusLDT5TtUlJKVPcxWOtBL6"
    "SBfa5qs7iYDHQOLwL14Ol9LJ33khCH7EN4xQtXVGo6vMTK2CnqSzv8fu6AoJmyg5Ra0xQ715"
    "giyZYFhsUdpdXDJC5ClOePetsgQ9KKo8WFQDkEKK5EfWCBcmYzmvDEP5qpwZMd+WGFdwrGUp"
    "yhQjmkilKeu+4yBNFNL5EFDLMka5HwlTq+4hMMpzVKIQOGwZOPaKMHI+1SuLIrp3VVXBAuAD"
    "wIGx2tcHAIHEGEuz1ay0AwaUQZCyv7vJxx5c4f/8v4kYw1mJv3OMOAS7oT7yloc3EI2jXm9/"
    "n9e89fEjNYD22n8otl/6bdfrXqOXX0cm00zPnaOV3sNouEnev85e9yIGSbOxRLO2QqIamNYW"
    "hS4pixHlyJBKny145s1rDCFej2rSiCdsEBJ0QZoo3P/X3pnFWHacdfxXVWe5++nb2/Ts4xnb"
    "OCExBGwJO4oQkEQRxiAHYiEQDshIgJCIeOMhCCSSB95AkYxixYAtIYF5iILjLNjOgnHsYMWJ"
    "sWPA42XGMx5PT09vdzlbnarioc69PTOKJ7TBzpjc/0svD33Vp+p82///fV+gfKpXWaJQEquK"
    "Sm9g9TatVkBWxYQtQek65KWlsvYiH5vl+dTcWmvQupyOjA+kIla+/CyVJAB/MepD9kGcZzfl"
    "ZGqXq7kB4320Uqq2eL7vRxIwHq8z3+lw7VXL/Pt/nEQqhXCmzgZEnRHUo/icqFfsCJ/uXRAI"
    "WgfGSD76B3ftKg34P18Zs/jOPxUvfecTrtF8Gr3xGsPNJlVcIKMOjd7V9IIlqAqKYpXB+GVK"
    "PUSJHu12j05jCSkDinwNa0u0HWFdgRJRXYWTvsGjHkJpHdjKP2RnfY9iJEMI3HQbeSQCKgvG"
    "lrRERisCbYeUIqByEWXUIrMOXRmqSoP1gZwufOXR1mKLKIrI0hSHq/cAtxiMhv6tr2sY1uy0"
    "wU/kMG5SGxCCKAjQuiSO6oGQFopsxCAbc+sH3sO3nnmJQAikCmpyzf8N/6UW2k4swbRUOfk0"
    "6Tew7hJvys6goz/+cfHoAx9zh+ZXmZsvaVRNyjIg33iOgd3Gxou0mov0Fq4jkA3INhmM1ljb"
    "eh7jSuLmMs1Gl3a8D1yILl/DygpTFZhKo6QvzggUURBSmqoenmDfLcgAAAyWSURBVCwRTk8P"
    "RQBIR1T/4Ooya2wNlaiwZgxsYtqKYRlQqZD1MVSlQ9eyLGstzUaDMAzQQYAtC1TdiIK1tdjU"
    "+2bf4ezqqakTnt5zEp1ma0osYS3OBlQ6Y7yVEknDNftXOHpomVfObPiCkVMo6ZCiAqdqevhi"
    "F+DsTiXQyYBhufuzetOWRr3v1r8UAN968FYXqyGdNiy2Y7riAMoIxsPzDDdfpcKh4gW6vT69"
    "hX1gYDw+xSg9x5Y+gZCGdnQtjaak2TAYk5OnnniqbIYm36FNhcFhkSLEUufKoiZVwJMszkfk"
    "Skio5wpJ50iikkpkJHOSohIUlWRjJKhsiFCKsigx1scpUtbmvz6PnXKtf62F9MUmYMo3hFFE"
    "ZXyl0zmLM6ZWQ/nA14zP8yu33MgnP/UAAoVUF4yEqVOJ148BQKqAvNhdFRDegq1hP3nLA1ND"
    "9eCnb3NHDhSYMGfvfJtu5ImkooD11ZPgLEEU0Yj2szR3yLuDwjAqn2Ftw4FRxEGLXtIlVAJL"
    "gq4KstEI5wyVLXGuIvRkLExq+dSq46l9dtMsIpC+FBUKgVD+kGLlMKGmHzuMLcnLlGERMTBQ"
    "CuXl5A6UCvzCq7p8OyGKjLVT6zwZMmnr3wnlWUwRCfI0n5JFVb7Fu665msP7F4GhL1XXmoRJ"
    "FHhp5/WF00JCGSB2IQad4C1dG3fL73z2ogDl0ftudUkvYqm3ynzPV+ysKRhUJ9g+VwIBURTS"
    "a1/PXAfAUOmC9fUT6NKiRECn1Sbp7kFgqXROnqe4aoQvMDE9GOcAK+qgrR5JY/2y5sAxXV4l"
    "padjrZTgJNY5YgXthmMlseSlptSCjSwiF45RztTdTOTck4zAOTfVIVS6qpVLvqFFlyVZnnkJ"
    "mVQgC/Jsi9tvvYl/+fKDU5WSmLgvgIssgKgHd9Q/SkV5JbmA/wned8cDF12Ix+9/v5vrCeZb"
    "AYtJA2kVRhvG609RaJBhj6idsLByLbFsUaZbpKOXOX3uHNYKoiim10voyiM4U5KXW5RmQJVl"
    "SOHfdmGNl6sKv6fI1aZcIpiOrQOk9Vs3FI5QgTVeqBGHFqcsvXAMDjIdkGvFuTQgKxy63uwp"
    "69pAFMXoqiJUXu4+mU1EZSjSnAD/vXIR1XiDH7vmKM8/m6AjL6oVRmBUgLGg2BnaKbFYaRFW"
    "oB2EKqYQu6QC+QFfgEtx0+0PTy/EQ/d8xC3PlySdkj1JRl82cGVOmWVsDF7G2Ii4uUCjc5Aj"
    "8z2srUizAVubm5zLHiOMYtrNZdqNFeJOTKXHpOMNtEmRwqDqOTyT92zSNwjUjKPf0zPpzZeq"
    "Fp0KcFLSVL4jJ1SWdqCZa2i0kRirWN02lC5mmFskIVMhh5sshHKUeTk13z7/10gTgE752fe+"
    "hzPHn0J2PSfgrPXy+AtcgE8LpY9rrEMIWXc67Q5X1AW4EB+48x8vsg5f/7vb3cJ8Trs1ZKHT"
    "RSrIizVG2xucLzZQqkm3tczelT0E8kcpy4zB6Axnzr+E1hndZpdub4mkuxdTpVTlGFdqr+2z"
    "GkHFpBtHTH2u/zopwXq4aUNKKBQoQeUkBI6o8hH/0UVHZSyFdpSuYlQoxjqisp7FLIoMa5kq"
    "mR3gRIXNC4xY5+qjB9g8fRyB8fIyWY+9vWD0jgSQE4lYnQbq3ZWB4Qq+AJfip3/9/umF+PKn"
    "3u/27GnQ78B8M6TfOopxhqxIOX36GUoMcdwlae/n0IEbcJSM03U2t9Z4rXgRJVp0Owt0OgkN"
    "KTHZEIpttNGYSoNXNvgPm4hQpimXzy4CCV7pLAkDibUFoVJY4YWiQRAQR17W3gtTKtOkMoKt"
    "HEYhDGyAqbuVvCu3BEJhbE6k4Kojx8jOH6fSFSqoY5cLXgnr/A5hT6H4nsvvsyPye2J35PEV"
    "iifu/ZBrtWMWuo5OSxPToywr0mybrBhi1DKt9iK9pE8YNrDFkOHWeTa31nBYmo0erfYCjUbk"
    "exzzESrdptLay92Eg6DAVQqJQjqJw3h+fzLY4tKn7yYqXrxWgZ0h0lhDTsQ4l4xLyerQYVyE"
    "kRKrAo7ug+TIzTz0D3dz7GCHogxpC78WevrnRYVAEkrD2Dg689fzwY/9/a7P821jAS6Hn/ro"
    "l6b/+Bfu+jW3d/kkzUZMv9Ol318kNCnj7FVWT59E24Ao6pL0+lx19QGchTQdMNg+y+q5TcJQ"
    "0G0nJN19xIGgKDLy0ZBIF/UcI41FE6gAVTe6BjiqHZsBUDN2O9qloB7y5ISAuukkahqSuOJA"
    "V1Lago1MsK5DirFhuaVZWtlDoQvfSynlJRoAEE5ha5n8ZLbAbvH/wgJcDl+95+fcSq9B0lR0"
    "moJGoBlpwVZaUpYpKlC04h5znX202nOko5LN7fNsj89gypJOp0W/26cjOjg0ZTYiS0f1/ouK"
    "wGqwFivU9/j0ndKNrMu0k4YZv5vLp5o+DnGUViMbMS4MiRb3s1a0efyRr3FoqU2a5551nEBa"
    "sKGfDBaElOIgv/zxB344LcDl8DN3PjJ9KA//9e2uF1cstyxJHBO3ezgnyLTmzOqLlKWh1WzT"
    "6SxyrH8DSsFosMXW+hqnyueQNqbXXSRZOELsDKYYo7MBlBlSa89UTkrAzk51AMKzRcBONdeX"
    "jSscgjhQiKokChSVggqDSVdZ2v9e4vYcRbaNk15h5CFqn18P3Q4UuphZgF3ja/d+2PVbAxba"
    "0AmahEFIkeWUhWaQpWgb0JpbZG5uiXbcpRxr0sEptjZPkJqYdrtHr7NAo9khLA1lOkDn20hT"
    "Yk1KIAIUAuEU2pY4KfDFaoiduEiO5eTOiDgpBDoMiI8c4+mXFS98+ykWW2O0i6YnZq0fGiGE"
    "wYVLnM0jfveTD+36PH+oL8ClePxvf9HNJxX9VkVTOKSISUcVw1FGYRxx3CeZW6LZ6hAHAVuD"
    "86yvrzJOt5DKMJ8sM9eeJw4aVNsZpR5RlFtYm9IyiklnuxCCaqJ4mvj1SVsd/gK4MCbttmnu"
    "uYHP3v85FjoV1lxI/wqU8rMWiJd4dRTw+3/+lZkL+N/gpt/8p+kD/Mbf/LyL4oDFOUe7HTMv"
    "QiqbszU8zpk13wiT9Pawd+UQrdb1jPMBw+11Tq2dIc2H9BtNkm6fbnIYYwLc+JRXH2vfHOuc"
    "XxQ11f5PVD9MlD6OSKe4YpX9hw9y/syLNJWbakKt9cpqa42X0L2BFBBmF+B1cfNvfWF6Gb54"
    "96+6Zcb0E0e/HbC33SFVbdLxgNNnn6MoSpKoT2duicMH3oGQDUbpaTa3Njm1dgZBTqd7hH5/"
    "iVYQo1ON2D4OWCrjB2I6IcGCxbeTyUoidEY2OsePvPPdnDrxXzQbimnnlJPT6qLCYvQbIAKY"
    "uYA3hM/f92F3Fdsk3Sbt0BIKx9gqRsMt8myMw9DuvJt+ktCMY4y2bI5eZH1jnTzNaHeazM1d"
    "RysOiTDk2RAGr4Hx086kM1TKEtHDdST53qt5+LFXYPgCTdME28IGAxRgnCPqH+OZk2P++NNf"
    "n8UAbzW+9Fe/5HodwUE1pNmKiSKFtZDqbYbDgryUBHGHheYyvSSpB0TnvLr2HNl4hLKCuU7C"
    "UrJCoEKs1ozTEa7cxJUF8UJE3ktYHe7j8UcfI4lLJLF/86XAOEeQHOa7J0b8yWcem8UAbzU+"
    "9Hufu+ihf/WeX3CN2HC02WZPv4t0FVYXrGenOL/xPNZZOu0mB/ZcR6PRwlQ5G+vrPL96gkpn"
    "NKKIpaUF+skxAuEY5UPE0HF4pcc3VQtjKywpiObUZZS6oiouvyX89TCzAG8ynrzvNtcLxyw2"
    "NEEYYZwkyzXD0ZhxoQlki4XFvfS7e5DSkY4HbKydZTvdhGhIq7HMfDNBLTf5z+Pw4rePoxqb"
    "QIyQeD6htZ87PrH7DABmFuBNx4137IhgHrn7I64fjVloluyfb4DskmkYjs/y7PnvYglJmiss"
    "rRzlUDNAmIL1tTXS4Wt0hOLw4Xfx9LMlC0TkRtCxAVqkbOrd9QJciJkF+AHi8btvcb2GYbGr"
    "SQgxxrGdp6yPhmSVI4z69OaWmVtaJHbbZMQ88Z2TnHvpBKGKaQhBpTSrJuEP/+LJN3SWswtw"
    "heAbn/mgi8KAhaZjMfb9CaWB7WHO9sDvUVo5vI+hnOcrD/0bbSdQIkBE8Era44/uemLmAt7O"
    "uPm3//miA3zynttcJyhY7ASsJG3ySlJVIxpJn0Z/AdZXPQHl1GXVwt8PswtwheLGO3dihy/e"
    "c5vrRyVLYUYvP8tPvOMgT3yzpFVuIlWPcdZ6w58zcwFvQ/zrvb/hXj23yurpc6xtjviz+16Y"
    "neMMM8wwwwwzzDDDDDPMMMMMM8wwwwyXx38D/HH6WAMkfKUAAAAASUVORK5CYII=")



import pytest

from pydicom.dataset import FileDataset, FileMetaDataset


@pytest.fixture(scope='function')
def dicom_generator(tmpdir):
    def _dicom(filename='image.dcm', **values):
        filename = str(tmpdir.join(filename))

        file_meta = FileMetaDataset()
        file_meta.MediaStorageSOPClassUID = '1.2.840.10008.5.1.4.1.1.2'
        file_meta.MediaStorageSOPInstanceUID = '1.2.3'
        file_meta.ImplementationClassUID = '1.2.3.4'

        ds = FileDataset(filename, {}, file_meta=file_meta, preamble=b'\0'*128)
        ds.is_little_endian = True
        ds.is_implicit_VR = False

        ds.PatientName = 'Jonathan^Suever'
        ds.SeriesDescription = 'Dicom Sort Test Series'
        ds.SeriesNumber = 1

        ds.update(values)

        ds.save_as(filename)

        return filename, ds

    return _dicom



from wx import PostEvent
from wx.lib.newevent import NewEvent

PathEvent, EVT_PATH = NewEvent()
PopulateEvent, EVT_POPULATE_FIELDS = NewEvent()
SortEvent, EVT_SORT = NewEvent()
CounterEvent, EVT_COUNTER = NewEvent()
UpdateEvent, EVT_UPDATE = NewEvent()

post_event = PostEvent



import os
import pydicom
import re
import sys

from pydicom.errors import InvalidDicomError

INVALID_FILENAME_CHARS = re.compile('[\\\\/\\:\\*\\?\\"\\<\\>\\|]+')

if sys.platform == 'win32':
    DIRECTORY_EXISTS_EXCEPTION = WindowsError
else:
    DIRECTORY_EXISTS_EXCEPTION = OSError


def mkdir(directory):
    try:
        os.makedirs(directory)
    except DIRECTORY_EXISTS_EXCEPTION:
        return


def recursive_replace_tokens(formatString, repobj):
    max_rep = 5
    rep = 0

    while re.search('%\\(.*\\)', formatString) and rep < max_rep:
        formatString = formatString % repobj
        rep = rep + 1

    return formatString


def clean_directory_name(path):
    return re.sub(INVALID_FILENAME_CHARS, '_', path)


def clean_path(path):
    outpath = ''

    head, tail = os.path.split(path)

    while tail:
        outpath = os.path.join(clean_directory_name(tail), outpath)
        head, tail = os.path.split(head)

    return os.path.join(head, outpath)[:-1]


def isdicom(filename):
    if os.path.basename(filename).lower() == 'dicomdir':
        return False
    try:
        return pydicom.read_file(filename)
    except InvalidDicomError:
        return False



import unittest
from dicomsort import config


class TestConfig(unittest.TestCase):
    def test_configuration_version(self):
        assert config.configuration_version == '2.0'

    def test_default_configuration(self):
        default = config.default_configuration

        # Version
        assert 'Version' in default
        assert default['Version'] == config.configuration_version

        # Anonymization
        assert 'Anonymization' in default
        anon = default['Anonymization']
        assert isinstance(anon, dict)

        assert 'Fields' in anon
        assert isinstance(anon['Fields'], list)

        assert 'Replacements' in anon
        assert isinstance(anon['Replacements'], dict)

        # Filename Format
        assert 'FilenameFormat' in default
        assert isinstance(default['FilenameFormat'], dict)

        # Miscellaneous Settings
        assert 'Miscpanel' in default
        assert isinstance(default['Miscpanel'], dict)



import configobj
import os
import sys
import wx

from dicomsort import config
from dicomsort import Metadata as meta
from dicomsort.dicomsorter import DicomSorter
from dicomsort.errors import DicomFolderError
from dicomsort.gui import errors, events, icons, preferences, widgets
from dicomsort.gui.dialogs import (
    AboutDlg, CrashReporter, HelpDlg, QuickRenameDlg, UpdateDlg
)
from dicomsort.gui.update import UpdateChecker

DEFAULT_FILENAME = '%(ImageType)s (%(InstanceNumber)04d)%(FileExtension)s'


def except_hook(exc_type, value, tb):
    dlg = CrashReporter(None, type=exc_type, value=value, traceback=tb)
    dlg.ShowModal()
    dlg.Destroy()


class DicomSort(wx.App):
    def __init__(self, *args):
        wx.App.__init__(self, *args)
        self.frame = MainFrame(None, -1, meta.pretty_name, size=(500, 500))
        self.SetTopWindow(self.frame)
        self.frame.Show()

        sys.excepthook = except_hook

        # Check for updates
        UpdateChecker(self.frame, listener=self.frame)
        self.frame.Bind(events.EVT_UPDATE, self.frame.OnNewVersion)

    def MainLoop(self, *args):
        wx.App.MainLoop(self, *args)


class MainFrame(wx.Frame):

    def __init__(self, *args, **kwargs):
        wx.Frame.__init__(self, *args, **kwargs)

        self.Create()
        self._InitializeMenus()

        # Use os.getcwd() for now
        self.dicom_sorter = DicomSorter()

        # Store the selected output directory to speed it up
        self.outputDirectory = None

        # Get config from parent
        self.config = configobj.ConfigObj(config.configuration_file)
        # Set interpolation to false since we use formatted strings
        self.config.interpolation = False

        # Check to see if we need to populate the config file
        if len(self.config.keys()) == 0:
            self.config.update(config.default_configuration)
            self.config.write()
        elif 'Version' not in self.config or \
                self.config['Version'] != config.configuration_version:
            self.config.update(config.default_configuration)
            self.config.write()

        self.prefDlg = preferences.PreferenceDlg(
            self, -1, '{} Preferences'.format(meta.pretty_name),
            config=self.config, size=(400, 400)
        )

        self.Bind(wx.EVT_CLOSE, self.OnQuit)

        self.CreateStatusBar()
        self.SetStatusText("Ready...")

    def OnNewVersion(self, evnt):
        dlg = UpdateDlg(self, evnt.version)
        dlg.Show()

    def Create(self):

        # Set Frame icon
        if os.path.isfile(os.path.join(sys.executable, 'DSicon.ico')):
            self.exedir = sys.executable
        else:
            self.exedir = sys.path[0]

        self.SetIcon(icons.main.GetIcon())

        vbox = wx.BoxSizer(wx.VERTICAL)

        self.pathEditor = widgets.PathEditCtrl(self, -1)
        vbox.Add(self.pathEditor, 0, wx.EXPAND)

        titles = ['DICOM Properties', 'Properties to Use']
        self.selector = widgets.FieldSelector(self, titles=titles)

        self.selector.Bind(events.EVT_SORT, self.Sort)

        vbox.Add(self.selector, 1, wx.EXPAND)

        self.SetSizer(vbox)

        self.pathEditor.Bind(events.EVT_PATH, self.FillList)

    def Sort(self, evnt):
        if self.dicom_sorter.is_sorting():
            return

        self.anonymize = evnt.anon

        miscTab = self.prefDlg.pages['Miscpanel']
        self.dicom_sorter.series_first = miscTab.seriesFirst.IsChecked()
        self.dicom_sorter.keep_original = miscTab.keepOriginal.IsChecked()

        if self.anonymize:
            anon_tab = self.prefDlg.pages['Anonymization']
            rules = anon_tab.anonList.GetAnonDict()
            self.dicom_sorter.set_anonymization_rules(rules)
        else:
            self.dicom_sorter.set_anonymization_rules(dict())

        directory_format = evnt.fields

        filename_method = int(self.config['FilenameFormat']['Selection'])
        filename_format = ''

        if filename_method == 0:
            # Image (0001)
            filename_format = DEFAULT_FILENAME
        elif filename_method == 1:
            # Use the original filename
            filename_format = ''
            self.dicom_sorter.keep_filename = True
            # Alternately we could pass None to dicom_sorter
        elif filename_method == 2:
            # Use custom format
            filename_format = self.config['FilenameFormat']['FilenameString']

        self.dicom_sorter.filename = filename_format
        self.dicom_sorter.folders = directory_format

        self.outputDirectory = self.SelectOutputDir()

        if not self.outputDirectory:
            return

        # Use for the real deal
        self.dicom_sorter.sort(self.outputDirectory, listener=self)

        self.Bind(events.EVT_COUNTER, self.OnCount)

    def OnCount(self, event):
        status = '%s / %s' % (event.Count, event.total)
        self.SetStatusText(status)

    def SelectOutputDir(self):
        if not self.outputDirectory:
            # Then don't set a default path
            dlg = wx.DirDialog(self, "Please select an output directory")
        else:
            dlg = wx.DirDialog(self, "Please select an output directory",
                               self.outputDirectory)

        dlg.CenterOnParent()

        if dlg.ShowModal() == wx.ID_OK:
            outputDir = dlg.GetPath()
        else:
            outputDir = None

        dlg.Destroy()

        return outputDir

    def OnPreferences(self, *_event):
        self.config = self.prefDlg.ShowModal()

    def OnQuit(self, *_event):
        sys.exit(0)

    def OnAbout(self, *_event):
        AboutDlg()

    def OnHelp(self, *_event):
        HelpDlg(self)

    def _MenuGenerator(self, parent, name, arguments):
        menu = wx.Menu()

        for item in arguments:
            if item == '----':
                menu.AppendSeparator()
            else:
                menuitem = wx.MenuItem(menu, -1, '\t'.join(item[0:2]))
                if item[2] != '':
                    self.Bind(wx.EVT_MENU, item[2], menuitem)

                menu.Append(menuitem)

        parent.Append(menu, name)

    def LoadDebug(self, *_event):
        size = self.Size
        pos = self.Position

        pos = (pos[0] + size[0], pos[1])

        self.debug = wx.Frame(
            self, -1, '{} DEBUGGER'.format(meta.pretty_name),
            size=(700, 500), pos=pos
        )

        self.crust = wx.py.crust.Crust(self.debug)
        self.debug.Show()

    def QuickRename(self, *_event):
        self.anonList = self.prefDlg.pages['Anonymization'].anonList
        dlg = QuickRenameDlg(
            self, -1, 'Anonymize', size=(250, 160), anonList=self.anonList
        )
        dlg.ShowModal()

    def _InitializeMenus(self):
        menubar = wx.MenuBar()

        file = [['&Open Directory', 'Ctrl+O', self.pathEditor.BrowsePaths],
                '----',
                ['&Preferences...', 'Ctrl+,', self.OnPreferences],
                '----',
                ['&Exit', 'Ctrl+W', self.OnQuit]]

        self._MenuGenerator(menubar, '&File', file)

        win = [['Quick &Rename', 'Ctrl+R', self.QuickRename], '----',
               ['&Debug Window', 'Ctrl+D', self.LoadDebug]]

        self._MenuGenerator(menubar, '&Window', win)

        help = [['About', '', self.OnAbout],
                ['&Help', 'Ctrl+?', self.OnHelp]]

        self._MenuGenerator(menubar, '&Help', help)

        self.SetMenuBar(menubar)

    def FillList(self, event):
        self.dicom_sorter.pathname = event.path
        try:
            fields = self.dicom_sorter.available_fields()
        except DicomFolderError:
            message = ''.join([';'.join(event.path), ' contains no DICOMs'])
            errors.throw_error(message, 'No DICOMs Present', parent=self)
            return

        self.selector.SetOptions(fields)

        # Now set seriesDescription as the default
        if len(self.selector.selected.Items) == 0:
            self.selector.selected.Append('SeriesDescription')

        # This is clunky
        # TODO: Change PrefDlg to a dict
        self.prefDlg.pages['Anonymization'].SetDicomFields(fields)

        self.Notify(events.PopulateEvent, fields=fields)

    def Notify(self, evntType, **kwargs):
        event = evntType(**kwargs)
        wx.PostEvent(self, event)



import unittest

from dicomsort.errors import DicomFolderError


class TestDicomFolderError(unittest.TestCase):
    def test_constructor(self):
        baseError = Exception('OH NO')
        err = DicomFolderError(baseError)

        assert isinstance(err, DicomFolderError)
        assert isinstance(err, Exception)

    def test_string_representation(self):
        baseError = Exception('OH NO')
        err = DicomFolderError(baseError)

        assert str(err) == repr(baseError)



import wx


class WxTestCase:
    def setup(self):
        self.app = wx.App()
        self.frame = wx.Frame(None)
        self.frame.Show()

    def teardown(self):
        def _cleanup():
            for win in wx.GetTopLevelWindows():
                if win:
                    if isinstance(win, wx.Dialog) and win.IsModal():
                        win.EndModal(0)
                    else:
                        win.Close(force=True)

            wx.WakeUpIdle()

        timer = wx.PyTimer(_cleanup)
        timer.Start(100)
        self.app.MainLoop()



import os
import unittest

from dicomsort import utils


class TestRecurvsiveReplaceTokens(unittest.TestCase):
    def test_no_tokens(self):
        format_string = 'no_token_string'
        output = utils.recursive_replace_tokens(format_string, {})

        assert output == format_string

    def test_non_nested_tokens(self):
        format_string = 'prefix_%(Key)s_suffix'
        values = {'Key': 'value'}
        output = utils.recursive_replace_tokens(format_string, values)

        assert output == 'prefix_value_suffix'

    def test_nested_tokens(self):
        format_string = 'prefix_%(Key1)s_suffix'
        values = {'Key1': '%(Key2)s', 'Key2': 'value'}
        output = utils.recursive_replace_tokens(format_string, values)

        assert output == 'prefix_value_suffix'

    def test_max_nested_tokens(self):
        format_string = 'prefix_%(Key1)s_suffix'
        # Make the values recurse
        values = {
            'Key1': '%(Key2)s',
            'Key2': '%(Key3)s',
            'Key3': '%(Key4)s',
            'Key4': '%(Key5)s',
            'Key5': '%(Key6)s',
            'Key6': 'blah',
        }

        # Will only perform replacements 5 times
        output = utils.recursive_replace_tokens(format_string, values)

        assert output == 'prefix_%(Key6)s_suffix'


class TestIsDicom:
    def test_dicomdir(self, dicom_generator):
        dicomdir, _ = dicom_generator('DICOMDIR')

        assert utils.isdicom(dicomdir) is False

    def test_invalid_dicom(self, tmpdir):
        fid = tmpdir.join('invalid')
        fid.write('invalid')

        assert utils.isdicom(str(fid)) is False

    def test_valid_dicom(self, dicom_generator):
        filename, _ = dicom_generator()

        assert utils.isdicom(filename) is not False


class TestCleanDirectoryName:
    def test_no_invalid_chars(self):
        dirname = 'clean'
        assert utils.clean_directory_name(dirname) == dirname

    def test_invalid_chars(self):
        dirname = 'prefix\\:*?|"<>suffix'
        assert utils.clean_directory_name(dirname) == 'prefix_suffix'


class TestCleanPath:
    def test_no_invalid_parts(self):
        path = os.path.join('/tmp/dir/ok')
        assert utils.clean_path(path) == path

    def test_with_invalid_parts(self):
        dirname = 'prefix:*?|"<>suffix'
        path = os.path.join('tmp', 'dir', dirname, dirname)
        expected = os.path.join('tmp', 'dir', 'prefix_suffix', 'prefix_suffix')
        assert utils.clean_path(path) == expected


class TestMkdir:
    def test_directory_exists(self, tmpdir):
        utils.mkdir(str(tmpdir))

    def test_new_directory(self, tmpdir):
        new_dir = str(tmpdir.join('new_dir'))
        utils.mkdir(new_dir)

        assert os.path.exists(new_dir)

    def test_new_nested_directory(self, tmpdir):
        new_dir = str(tmpdir.join('new_dir').join('nested_dir'))
        utils.mkdir(new_dir)

        assert os.path.exists(new_dir)



import os
import re
import wx

import wx.grid
import wx.html

from wx.lib.mixins.listctrl import ListCtrlAutoWidthMixin, TextEditMixin

from dicomsort.gui import errors, events
from dicomsort.gui.dialogs import SeriesRemoveWarningDlg


class FileDropTarget(wx.FileDropTarget):
    def __init__(self, callback):
        super(FileDropTarget, self).__init__()
        self.callback = callback

    def OnDropFiles(self, x, y, filenames):
        self.callback(x, y, filenames)


class CustomDataTable(wx.grid.GridTableBase):
    def __init__(self, data):
        super(CustomDataTable, self).__init__()

        self.colLabels = ['', 'DICOM Property', 'Replacement Value']

        self.dataTypes = [
            wx.grid.GRID_VALUE_BOOL,
            wx.grid.GRID_VALUE_STRING,
            wx.grid.GRID_VALUE_STRING
        ]

        if data is None:
            data = [['', '', ''], ]

        self.data = data

    # --------------------------------------------------
    # required methods for the wxPyGridTableBase interface

    def GetNumberRows(self):
        return len(self.data)

    def GetNumberCols(self):
        return len(self.data[0])

    def IsEmptyCell(self, row, col):
        try:
            return not self.data[row][col]
        except IndexError:
            return True

    # Get/Set values in the table.  The Python version of these
    # methods can handle any data-type, (as long as the Editor and
    # Renderer understands the type too,) not just strings as in the
    # C++ version.
    def GetValue(self, row, col):
        try:
            return self.data[row][col]
        except IndexError:
            return ''

    def SetValue(self, row, col, value):
        def innerSetValue(row, col, value):
            try:
                self.data[row][col] = value
            except IndexError:
                # add a new row
                self.data.append([''] * self.GetNumberCols())
                innerSetValue(row, col, value)

                # tell the grid we've added a row
                msg = wx.grid.GridTableMessage(
                    self,
                    wx.grid.GRIDTABLE_NOTIFY_ROWS_APPENDED,
                    1
                )

                view = self.GetView()

                if view:
                    view.ProcessTableMessage(msg)
        innerSetValue(row, col, value)

    # --------------------------------------------------
    # Some optional methods

    # Called when the grid needs to display labels
    def GetColLabelValue(self, col):
        return self.colLabels[col]

    # Called to determine the kind of editor/renderer to use by
    # default, doesn't necessarily have to be the same type used
    # natively by the editor/renderer if they know how to convert.
    def GetTypeName(self, row, col):
        return self.dataTypes[col]

    # Called to determine how the data can be fetched and stored by the
    # editor and renderer.  This allows you to enforce some type-safety
    # in the grid.
    def CanGetValueAs(self, row, col, typeName):
        colType = self.dataTypes[col].split(':')[0]
        if typeName == colType:
            return True
        else:
            return False

    def CanSetValueAs(self, row, col, typeName):
        return self.CanGetValueAs(row, col, typeName)


class CheckListCtrl(wx.ListCtrl, ListCtrlAutoWidthMixin, TextEditMixin):

    def __init__(self, parent):
        wx.ListCtrl.__init__(self, parent, -1, style=wx.LC_REPORT)
        ListCtrlAutoWidthMixin.__init__(self)

        # The order here matter because we favor Check over Edit
        TextEditMixin.__init__(self)

        self.editColumns = []

        self.EnableCheckBoxes()

    def _GetCheckedIndexes(self):
        return [i for i in range(self.ItemCount) if self.IsItemChecked(i)]

    def ClearColumn(self, col):
        [self.SetItem(i, col, '') for i in range(self.ItemCount)]

    def SetColumnEditable(self, column, edit=True):
        if edit:
            if column not in self.editColumns:
                self.editColumns.append(column)
        else:
            if column in self.editColumns:
                self.editColumns.remove(column)

    def SetStringItems(self, items):
        self.DeleteAllItems()

        for item in items:
            if isinstance(item, str):
                item = [item, ]

            row = self.InsertItem(self.GetItemCount() + 1, item[0])

            for col in range(1, len(item)):
                self.SetStringItem(row, col, item[col])

    def CheckItems(self, itemIndex):
        [self.CheckItem(index) for index in itemIndex]

    def GetCheckedItems(self, col=None):
        return [self.GetItemList(r, col) for r in self._GetCheckedIndexes()]

    def UnCheckAll(self):
        [self.CheckItem(index, False) for index in range(self.ItemCount)]

    def GetCheckedStrings(self, col=None):
        return [self.GetStringItem(r, col) for r in self._GetCheckedIndexes()]

    def FindStrings(self, strings, col=0):
        fields = [item.Text for item in self.GetItemList(col)]

        indices = list()

        for string in strings:
            try:
                indices.append(fields.index(string))
            except ValueError:
                indices.append(None)

        return indices

    def GetItemList(self, column=None):
        if column is None:
            return [self.GetItemList(c) for c in range(self.ColumnCount)]
        else:
            return [self.GetItem(r, column) for r in range(self.ItemCount)]

    def GetStringItem(self, row, column=None):
        if column is None:
            return [
                self.GetStringItem(row, c) for c in range(self.ColumnCount)
            ]
        else:
            return self.GetItem(row, column).Text

    def OpenEditor(self, column, row):
        """
        Overloaded method to render certain columns un-editable
        """
        if column not in self.editColumns:
            return
        else:
            # Call superclass edit method
            TextEditMixin.OpenEditor(self, column, row)


class PathEditCtrl(wx.Panel):
    def __init__(self, *args, **kwargs):
        wx.Panel.__init__(self, *args, **kwargs)
        self.create()
        self.path = ''
        self.Bind(wx.EVT_TEXT_ENTER, self.ValidatePath, self.edit)

        dt = FileDropTarget(self.ValidateDropFiles)
        self.edit.SetDropTarget(dt)

    def ValidateDropFiles(self, x, y, filenames):
        self.SetPaths(filenames)

    def ValidatePath(self, *evnt):
        paths = self.edit.GetValue()

        self.SetPaths(paths.split(';'))

    def create(self):
        hbox = wx.BoxSizer(wx.HORIZONTAL)

        opts = wx.ALIGN_CENTER_VERTICAL | wx.LEFT | wx.RIGHT | wx.TOP

        self.edit = wx.TextCtrl(self, -1, "", style=wx.TE_PROCESS_ENTER)
        hbox.Add(self.edit, 1, opts, 5)

        self.browse = wx.Button(self, -1, label="Browse")
        hbox.Add(self.browse, 0, opts, 10)

        self.browse.Bind(wx.EVT_BUTTON, self.BrowsePaths)

        self.SetSizer(hbox)

    def SetPaths(self, paths):
        if not isinstance(paths, list):
            paths = [paths, ]

        # Determine if there are any invalid paths
        badPaths = [path for path in paths if not os.path.isdir(path)]

        if len(badPaths):
            p = ', '.join(badPaths)
            errorMsg = 'The Following directories are invalid paths: %s' % p
            errors.throw_error(errorMsg, 'Invalid Paths', parent=self.Parent)
            return

        self.path = paths
        self.edit.SetValue(';'.join(paths))

        # Trigger an EVT_PATH event
        self.Notify()

    def Notify(self, *evnt):
        event = events.PathEvent(path=self.path)
        wx.PostEvent(self, event)

    def BrowsePaths(self, *evnt):
        # Open the dialog and search
        if len(self.path):
            defaultPath = self.path[0]
        else:
            defaultPath = ''

        pathDlg = wx.DirDialog(self.GetParent(), "Please Select Directory",
                               defaultPath)
        pathDlg.CenterOnParent()

        if pathDlg.ShowModal() == wx.ID_OK:
            path = pathDlg.GetPath()
            self.SetPaths([path, ])

        pathDlg.Destroy()


class FieldSelector(wx.Panel):

    def __init__(self, parent, id=-1, size=(250, 300), choices=None,
                 titles=None):
        wx.Panel.__init__(self, parent, id=id, size=size)

        self.choices = choices or []
        self.titles = titles or ['', '']

        self.create()

    def Filter(self, string=None):
        if string and isinstance(string, str):
            self.search.SetValue(string)
        else:
            string = self.search.GetValue()

        newVals = [p for p in self.choices if re.search(string, p,
                                                        re.IGNORECASE)]

        self.options.SetItems(newVals)

    def _return_focus(self):
        return

    def _sort_callback(self, *evnt):
        event = events.SortEvent(
            anon=self.anonQ.IsChecked(), fields=self.GetFormatFields()
        )
        wx.PostEvent(self, event)

    def GetSelectedItems(self):
        return self.selected.GetItems()

    def GetFormatFields(self):
        items = self.GetSelectedItems()
        return ['%(' + item + ')s' for item in items]

    def _anon_tick(self, args):
        if self.anonQ.IsChecked():
            self.Parent.QuickRename()

    def create(self):

        self.titleL = wx.StaticText(self, -1, self.titles[0])
        self.options = wx.ListBox(self, -1, choices=self.choices)

        self.search = wx.SearchCtrl(
            self, -1, "", style=wx.TE_PROCESS_ENTER)

        self.titleR = wx.StaticText(self, -1, self.titles[1])
        self.selected = wx.ListBox(self, -1, choices=[])

        self.sortBtn = wx.Button(self, -1, label="Sort Images")
        self.anonQ = wx.CheckBox(self, -1, label="Anonymize Data")

        # Make it so that when we check the checkbox it brings up the quick
        # rename dialog
        self.anonQ.Bind(wx.EVT_CHECKBOX, self._anon_tick)

        self.sortBtn.Bind(wx.EVT_BUTTON, self._sort_callback)

        # Setup double-click callbacks
        self.options.Bind(wx.EVT_LISTBOX_DCLICK, self.SelectItem)
        self.selected.Bind(wx.EVT_LISTBOX_DCLICK, self.DeselectItem)

        # Setup Search Control
        self.search.Bind(wx.EVT_TEXT, self.Filter)
        self.search.Bind(wx.EVT_TEXT_ENTER, self._return_focus)

        # Setup controls:
        self.bAdd = wx.Button(self, -1, label=">>", size=(50, -1))
        self.bRemove = wx.Button(self, -1, label="<<", size=(50, -1))
        self.bUp = wx.Button(self, -1, label="Up", size=(50, -1))
        self.bDown = wx.Button(self, -1, label="Down", size=(50, -1))

        self.bAdd.Bind(wx.EVT_BUTTON, self.SelectItem)
        self.bRemove.Bind(wx.EVT_BUTTON, self.DeselectItem)
        self.bUp.Bind(wx.EVT_BUTTON, self.PromoteSelection)
        self.bDown.Bind(wx.EVT_BUTTON, self.DemoteSelection)

        self._initialize_layout()

    def WidgetList(self):
        a = [self.options, self.search, self.selected,
             self.bAdd, self.bRemove, self.bUp, self.bDown]
        return a

    def DisableAll(self):
        [item.Disable() for item in self.WidgetList()]

    def EnableAll(self):
        [item.Enable(True) for item in self.WidgetList()]

    def _initialize_layout(self):
        # BoxSizer containing the options to select
        vboxOptions = wx.BoxSizer(wx.VERTICAL)

        vboxOptions.Add(self.titleL, 0, wx.ALIGN_CENTER_HORIZONTAL)
        vboxOptions.Add(self.options, 1,
                        wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, 10)
        vboxOptions.Add(self.search, 0, wx.EXPAND | wx.ALL, 10)

        # BoxSizer containing the selected options
        vboxSelect = wx.BoxSizer(wx.VERTICAL)

        vboxSelect.Add(self.titleR, 0, wx.ALIGN_CENTER_HORIZONTAL)
        vboxSelect.Add(self.selected, 1,
                       wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, 10)
        vboxSelect.Add(self.anonQ, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.ALL, 5)
        vboxSelect.Add(
            self.sortBtn, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.ALL, 10)

        # BoxSizer housing the controls
        vboxControl = wx.BoxSizer(wx.VERTICAL)

        vboxControl.Add(self.bAdd, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.TOP, 30)
        vboxControl.Add(
            self.bRemove, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.TOP, 15)
        vboxControl.Add(self.bUp, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.TOP, 15)
        vboxControl.Add(self.bDown, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.TOP, 15)

        # Horizontal Box that contains all elements
        hbox = wx.BoxSizer(wx.HORIZONTAL)

        hbox.Add(vboxOptions, 1, wx.EXPAND | wx.ALL, 10)
        hbox.Add(vboxControl, 0, wx.ALL, 0)
        hbox.Add(vboxSelect, 1, wx.EXPAND | wx.ALL, 10)

        self.SetSizer(hbox)

    def has_default(self):
        if self.selected.GetCount() == 0:
            return False

        if self.selected.GetItems()[-1] == 'SeriesDescription':
            return True
        else:
            return False

    def PromoteSelection(self, *evnt):
        self._move_selection(-1)

    def DemoteSelection(self, *evnt):
        self._move_selection(1)

    def _move_selection(self, inc):
        """
        Shared method for moving an item around the list.
        There really has to be a better way...
        """

        index = self.selected.GetSelection()

        if index == self.selected.GetCount() - 1 and self.has_default():
            return

        if inc < 0 and (self.selected.Count == 1 or index == 0):
            return
        elif (inc > 0 and index == self.selected.Count - 1):
            return

        prop = self.selected.GetStringSelection()

        self.selected.Delete(index)
        self.selected.Insert(prop, index + inc)
        self.selected.Select(index + inc)

    def SelectItem(self, *evnt):
        item = self.options.GetStringSelection()
        if self.has_default():
            self.selected.Insert(item, self.selected.GetCount() - 1)
        else:
            self.selected.Append(item)

    def DeselectItem(self, *evnt):
        index = self.selected.GetSelection()

        if index == -1:
            return

        if index == self.selected.GetCount() - 1 and self.has_default():
            warn = SeriesRemoveWarningDlg(None)
            warn.ShowModal()

            if warn.choice == 0:
                return
            elif warn.choice == 1:
                # change to original filenames
                cfg = self.GetParent().config
                cfg['FilenameFormat']['Selection'] = 1
                page = self.GetParent().prefDlg.pages['FilenameFormat']
                page.UpdateFromConfig(cfg)

        self.selected.Delete(index)

        self.selected.Select(index - 1)

    def SetTitles(self, titleL, titleR):
        self.leftBox

    def SetOptions(self, optionList=[]):
        # clear
        self.options.SetItems(optionList)
        self.choices = optionList



import os

from typing import Any, Dict, List

__version__ = '3.0.0'


# Project metadata
class Metadata:
    name: str = 'dicomsort'
    pretty_name: str = 'DICOM Sort'
    description: str = 'A DICOM Sorting Utility'
    license: str = 'MIT'
    version: str = __version__
    copyright: str = 'Copyright 2011 - 2021, Jonathan Suever'

    repository: str = 'https://github.com/dicomsort/dicomsort'
    website: str = 'https://dicomsort.com'
    download_url: str = website + '/downloads.html'
    issue_url: str = repository + '/issues/new'
    version_url: str = website + '/current'

    author: str = 'Jonathan Suever'
    author_email: str = 'suever@gmail.com'
    maintainer: str = author
    maintainer_email: str = author_email

    keywords: List[str] = [
        'dicom',
        'sorting',
        'images',
    ]

    classifiers: List[str] = [
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Science/Research',
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Software Development :: Libraries'
    ]

    @classmethod
    def readme(cls) -> str:
        d = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
        with open(os.path.join(d, 'README.md')) as fid:
            return fid.read()

    @classmethod
    def to_dict(cls) -> Dict[str, Any]:
        return {
            'name': cls.name,
            'version': cls.version,
            'description': cls.description,
            'long_description': cls.readme(),
            'long_description_content_type': 'text/markdown',
            'keywords': ' '.join(cls.keywords),

            'license': cls.license,
            'classifiers': cls.classifiers,

            'author': cls.author,
            'author_email': cls.author_email,
            'maintainer': cls.maintainer,
            'maintainer_email': cls.maintainer_email,

            'url': cls.website,
            'download_url': cls.download_url,
        }



import wx
import wx.py
import configobj

from dicomsort import config
from dicomsort.gui.anonymizer import AnonymizeList


class PreferencePanel(wx.Panel):

    def __init__(self, parent, shortname, title, config):
        wx.Panel.__init__(self, parent, -1)
        self.shortname = shortname
        self.config = config
        self.title = title

    def GetState(self):
        raise TypeError('Abstract Method!')

    def SaveState(self, *evnt):
        # Load since last saved version
        tmpconfig = configobj.ConfigObj(self.config.filename)
        self.StoreState(config=tmpconfig)

        # Write just the change from this panel
        tmpconfig.write()

    def UpdateFromConfig(self, config=None):
        raise TypeError('Abstract Method!')

    def RevertState(self, *evnt):
        # Load a temporary copy
        tmpconfig = configobj.ConfigObj(self.config.filename)
        self.config[self.shortname] = tmpconfig[self.shortname]
        # TODO: See if all of these configobjs need to be closed()

    def StoreState(self, config=None):
        if config is None:
            config = self.config

        config[self.shortname] = self.GetState()


class MiscPanel(PreferencePanel):
    def __init__(self, parent, config):
        super(MiscPanel, self).__init__(parent, 'Miscpanel',
                                        'Miscellaneous', config)
        self.create()

    def GetState(self):
        return {
            'KeepSeries': str(self.keepSeries.IsChecked()),
            'SeriesFirst': str(self.seriesFirst.IsChecked()),
            'KeepOriginal': str(self.keepOriginal.IsChecked())
        }

    def RevertState(self, *evnt):
        super(MiscPanel, self).RevertState()
        savedConfig = configobj.ConfigObj(self.config.filename)
        self.UpdateFromConfig(savedConfig)

    def UpdateFromConfig(self, config):
        tmp = config[self.shortname]
        if 'KeepSeries' not in tmp:
            tmp['KeepSeries'] = 'True'

        if 'SeriesFirst' not in tmp:
            tmp['SeriesFirst'] = 'False'

        if 'KeepOriginal' not in tmp:
            tmp['KeepOriginal'] = 'True'

        if tmp['KeepSeries']:
            self.keepSeries.SetValue(eval(tmp['KeepSeries']))

        if tmp['SeriesFirst']:
            self.seriesFirst.SetValue(eval(tmp['SeriesFirst']))

        if tmp['KeepOriginal']:
            self.keepOriginal.SetValue(eval(tmp['KeepOriginal']))

    def create(self):
        vbox = wx.BoxSizer(wx.VERTICAL)

        self.keepSeries = wx.CheckBox(
            self, -1,
            'Automatically include SeriesDescription'
        )

        # Vertical spacing
        vbox.Add((30, 40), 0, wx.ALL | 30)
        vbox.Add(self.keepSeries, 0, wx.ALIGN_LEFT | wx.ALL | 15)

        self.seriesFirst = wx.CheckBox(self, -1, "Series Name First")

        # Vertical spacing
        vbox.Add((30, 20), 0, wx.ALL | 30)
        vbox.Add(self.seriesFirst, 0, wx.ALIGN_LEFT | wx.ALL | 15)

        self.keepOriginal = wx.CheckBox(self, -1, "Keep Original Files")

        vbox.Add((30, 20), 0, wx.ALL | 30)
        vbox.Add(self.keepOriginal, 0, wx.ALIGN_LEFT | wx.ALL | 15)

        hbox = wx.BoxSizer(wx.HORIZONTAL)
        self.store = wx.Button(self, -1, "Set as Default", size=(120, -1))
        self.revert = wx.Button(self, -1, "Restore Default", size=(120, -1))
        self.Bind(wx.EVT_BUTTON, self.RevertState, self.revert)
        self.Bind(wx.EVT_BUTTON, self.SaveState, self.store)

        hbox.Add(self.store, 0, wx.ALL, 5)
        hbox.Add(self.revert, 0, wx.ALL, 5)

        vbox.Add(hbox, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.TOP, 25)

        self.SetSizer(vbox)


class FileNamePanel(PreferencePanel):

    def __init__(self, parent, config):
        PreferencePanel.__init__(self, parent, 'FilenameFormat',
                                 'Filename Format', config)

        self.create()

    def create(self):
        vbox = wx.BoxSizer(wx.VERTICAL)

        boxChoices = [
            'Image (0001)',
            'Keep Original Filename',
            'Custom Filename Format:'
        ]

        self.radioBox = wx.RadioBox(self, -1,
                                    style=wx.VERTICAL | wx.BORDER_NONE,
                                    choices=boxChoices)

        self.Bind(wx.EVT_RADIOBOX, self.OnChange, self.radioBox)

        vbox.Add(self.radioBox, 0, wx.TOP | wx.LEFT, 15)

        self.custom = wx.TextCtrl(self, -1, '',
                                  size=(300, -1))
        vbox.Add(self.custom, 0, wx.LEFT, 50)

        hbox = wx.BoxSizer(wx.HORIZONTAL)
        self.store = wx.Button(self, -1, "Set as Default", size=(120, -1))
        self.revert = wx.Button(self, -1, "Restore Default", size=(120, -1))
        self.Bind(wx.EVT_BUTTON, self.RevertState, self.revert)
        self.Bind(wx.EVT_BUTTON, self.SaveState, self.store)

        hbox.Add(self.store, 0, wx.ALL, 5)
        hbox.Add(self.revert, 0, wx.ALL, 5)

        vbox.Add(hbox, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.TOP, 25)

        self.SetSizer(vbox)

    def RevertState(self, *evnt):
        super(FileNamePanel, self).RevertState()
        savedConfig = configobj.ConfigObj(self.config.filename)
        self.UpdateFromConfig(savedConfig)

    def OnChange(self, *evnt):
        index = self.radioBox.GetSelection()

        # TODO: Make this more robust in the future
        if index != 2:
            self.custom.Disable()
        else:
            self.custom.Enable()

    def UpdateFromConfig(self, config):

        config.interpolation = False

        data = config[self.shortname]

        if 'Selection' in data:
            self.radioBox.SetSelection(int(data['Selection']))
        else:
            self.radioBox.SetSelection(0)

        self.custom.SetValue(data['FilenameString'])

        self.OnChange()

    def GetState(self):
        return {
            'FilenameString': self.custom.GetValue(),
            'Selection': self.radioBox.GetSelection()
        }


class PreferenceDlg(wx.Dialog):

    def __init__(self, *args, **kwargs):
        if 'config' in kwargs:
            self.config = kwargs.pop('config')
        else:
            self.config = configobj.ConfigObj(config.configuration_file)

        wx.Dialog.__init__(self, *args, **kwargs)

        # We want to make this a dict instead of list
        self.pages = dict()
        self.create()

        # Initialize from Config on Create
        self.UpdateFromConfig(self.config)

    def UpdateFromConfig(self, config=None):
        if config is None:
            config = self.config

        [val.UpdateFromConfig(config) for val in self.pages.values()]

    def Show(self, *args):
        # Call superclass constructor
        wx.Dialog.Show(self, *args)

    def create(self):
        self.nb = wx.Notebook(self)
        self.AddModule(AnonymousPanel)
        self.AddModule(FileNamePanel)
        self.AddModule(MiscPanel)

        vbox = wx.BoxSizer(wx.VERTICAL)
        vbox.Add(self.nb, 1, wx.EXPAND)

        hbox = wx.BoxSizer(wx.HORIZONTAL)

        self.cancel = wx.Button(self, -1, 'Cancel')
        self.apply = wx.Button(self, -1, 'Apply')

        self.cancel.Bind(wx.EVT_BUTTON, self.OnCancel)
        self.apply.Bind(wx.EVT_BUTTON, self.OnApply)

        hbox.Add(self.cancel, 0, 0, 10)
        hbox.Add(self.apply, 0, wx.LEFT, 10)

        vbox.Add(hbox, 0, wx.ALIGN_RIGHT | wx.TOP, 5)
        vbox.Add((10, 10), 0, wx.ALL | 30)

        self.SetSizer(vbox)

    def AddModule(self, panel):
        newPage = panel(self.nb, self.config)

        self.pages[newPage.shortname] = newPage
        self.nb.AddPage(newPage, newPage.title)

    def OnApply(self, *evnt):
        [val.StoreState() for val in self.pages.values()]

        # For now don't write this to the file because it's a local default
        self.Close()
        return self.config

    def OnCancel(self, *evnt):
        # Return configobj that we came in with
        self.Close()
        self.UpdateFromConfig(self.config)
        return self.config

    def ShowModal(self, *args):
        wx.Dialog.ShowModal(self, *args)
        return self.config


class AnonymousPanel(PreferencePanel):

    def __init__(self, parent, config):
        super(AnonymousPanel, self).__init__(parent, 'Anonymization',
                                             'Anonymizing Fields', config)

        self.create()

    def GetState(self):
        # Get all fields that are stored in config but not present in current
        defFields = self.config[self.shortname]['Fields']

        fields = list()
        # Keep only the empty ones
        for i, val in enumerate(self.anonList.FindStrings(defFields)):
            if val is None:
                fields.append(defFields[i])

        # Add to this list the newly checked ones
        fields.extend(self.anonList.GetCheckedStrings(0))

        return {
            'Fields': fields,
            'Replacements': self.anonList.GetReplacementDict()
        }

    def RevertState(self, *evnt):
        # Update self.config
        super(AnonymousPanel, self).RevertState()
        savedConfig = configobj.ConfigObj(self.config.filename)
        savedConfig.interpolation = False
        self.UpdateFromConfig(savedConfig)

    def SetDicomFields(self, values):
        self.anonList.SetStringItems(values)
        self.UpdateFromConfig(self.config)

    def UpdateFromConfig(self, config):
        data = config[self.shortname]

        # The fields that we care about are "Fields" and "Replacements"
        fields = data['Fields']
        self.anonList.UnCheckAll()
        self.anonList.CheckStrings(fields, col=0)
        self.anonList.ClearColumn(1)

        # Now put in substitutes
        self.anonList.SetReplacementDict(data['Replacements'])

    def create(self):
        vbox = wx.BoxSizer(wx.VERTICAL)

        title = wx.StaticText(self, -1, "Fields to Omit")
        vbox.Add(title, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.TOP | wx.BOTTOM, 10)

        self.anonList = AnonymizeList(self)

        hbox = wx.BoxSizer(wx.HORIZONTAL)

        self.store = wx.Button(self, -1, "Set as Default", size=(120, -1))
        self.revert = wx.Button(self, -1, "Restore Default", size=(120, -1))
        self.revert.Bind(wx.EVT_BUTTON, self.RevertState)
        self.store.Bind(wx.EVT_BUTTON, self.SaveState)

        hbox.Add(self.store, 0, wx.TOP | wx.LEFT, 10)
        hbox.Add(self.revert, 0, wx.TOP | wx.LEFT, 10)

        vbox.Add(self.anonList, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, 10)
        vbox.Add(hbox, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.BOTTOM, 15)
        self.SetSizer(vbox)



from dicomsort import Metadata as meta

helpHTML = """
<span style="fontsize:9px;"> <center><h3>{} Help</h3></center> <br>
<strong><u>Introduction</u></strong><br> This utility will sort images based
upon DICOM header information.  The user is able to sort by multiple fields
which are sent to the program when a directory of DICOM images is selected.
<br><br> <strong><u>Operation</u></strong><br> You must first Browse for the
directory that contains the DICOM images to be sorted. <strong>This folder
should preferably be the DICOM folder on the CD</strong> <br><br> When the
images are loaded the available DICOM header values are listed in the
<strong>DICOM Properties</strong> listbox. Double-click or push the
<strong>&gt;&gt;</strong> button to add the selected property to the listbox on
the right.  <br> The properties that are added to the listbox on the right are
then used to determine the directory structure as shown below: <br><br>
<strong>/First Listbox Entry/Second Listbox Entry/ (etc.)</strong> <br><br> It
is important to note that you will select the output directory (signified by
'~' above) when you click on the <strong>Sort Button</strong>.  <br><br> You
can also change the ordering of the sorting parameters using the <strong>Up and
Down Buttons</strong> located between the two boxes.  <br><br> To remove a
property from the parameter box, use the <strong>&lt;&lt;</strong> button after
selecting the property to remove or double-click the entry.  Once you are
satisfied with your parameter selection, click on the <strong>Sort
Button</strong> to perform the requested operation.  <br><br> Similarly to
previous versions of the sorting program, the program will always additionally
sort the data by <strong>Series Number</strong> and <strong>Series
Description</strong>.  The actual image file name is determined by the
<strong>Instance Number</strong> and <strong>Image Type</strong> (Magnitude,
Phase, or Other).  <br><br> The number of images processed is shown in the
lower left-hand corner of the application.  <br><br> <strong><u>Optional
Features</u></strong><br> Version 1.1 now provides the user the ability to
remove sensitive patient data from the DICOM metadata. To do this, select the
<strong>Anonymize</strong> checkbox.  Pressing this button will bring up a
window that will allow you to select which fields you do not want included in
the output images. You can always revert back to the default properties using
the <strong>Restore Default</strong> button.  <br><br> If you wish to change
the properties that are considered to be private by default, you can edit the
anon.inc file located in the application directory. Just edit this file using
any text editor and make sure to include one field per line in the file. The
next time the program is loaded, these fields will be used as the default
anonymous fields.  <br><br> Another option is to use the original filenames as
the input files rather than renaming based upon the image type. This plays an
important role if the analysis program to be used requires a certain naming
convention. Names will be preserved when the checkbox is ticked.  <br><br>
<strong><u>Future Directions</u></strong> <ul> <li>Be able to handle selected
folders that contain more images than just DICOM images (create an
exception).</li> <li>Implement a preview of the values of selected DICOM
properties</li> </ul>
""".format(meta.pretty_name)
